from types import ModuleType
from typing import Any, Optional, TypeVar

T = TypeVar("T")
R = TypeVar("R")


class LazyLoaderMeta(type):
    """
    Meta class for lazy loading attributes from a module.
    """

    def __getattr__(cls, name: str) -> Any:
        module = cls._load()
        try:
            return getattr(module, name)
        except AttributeError as e:
            raise e

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        module = cls._load()
        return module(*args, **kwargs)


class router(metaclass=LazyLoaderMeta):
    """
    Lazy loader for a module.
    Used internally in the zyx library.

    Refer to __init__.py files.
    """

    import threading

    _module_cache: Optional[ModuleType] = None
    _lock = threading.Lock()

    @classmethod
    def _load(cls) -> Any:
        import inspect
        from importlib import import_module

        if cls._module_cache is None:
            with cls._lock:
                if cls._module_cache is None:
                    try:
                        module = import_module(cls._module_name)
                        cls._module_cache = getattr(module, cls._attr_name)

                        if inspect.isclass(cls._module_cache):
                            return type(cls.__name__, (cls._module_cache,), {})
                        elif inspect.isgeneratorfunction(cls._module_cache):
                            return cls._wrap_generator(cls._module_cache)
                        elif inspect.isfunction(cls._module_cache):
                            return cls._wrap_function(cls._module_cache)
                    except (ImportError, AttributeError) as e:
                        raise e
        return cls._module_cache

    @staticmethod
    def _wrap_generator(gen_func):
        def wrapper(*args, **kwargs):
            return gen_func(*args, **kwargs)

        return wrapper

    @staticmethod
    def _wrap_function(func):
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    @classmethod
    def init(cls, module_name: str, attr_name: str) -> None:
        cls._module_name = module_name
        cls._attr_name = attr_name

    def __getattr__(self, name: str) -> Any:
        module = self._load()
        try:
            return getattr(module, name)
        except AttributeError as e:
            raise e

    def __iter__(self):
        module = self._load()
        if hasattr(module, "__iter__"):
            return iter(module)
        raise TypeError(f"{module} object is not iterable")

    def __next__(self):
        module = self._load()
        if hasattr(module, "__next__"):
            return next(module)
        raise TypeError(f"{module} object is not an iterator")

    def __getitem__(self, item):
        module = self._load()
        if hasattr(module, "__getitem__"):
            return module[item]
        raise TypeError(f"{module} object does not support indexing")

    def __len__(self):
        module = self._load()
        if hasattr(module, "__len__"):
            return len(module)
        raise TypeError(f"{module} object does not have a length")

    def __contains__(self, item):
        module = self._load()
        if hasattr(module, "__contains__"):
            return item in module
        raise TypeError(f"{module} object does not support membership test")



# message parameter type


import sys

if sys.version_info >= (3, 11):
    from typing import (
        Any,
        Literal,
        Mapping,
        NotRequired,
        Sequence,
        TypedDict,
        Union,
        List,
    )
else:
    from typing import Any, Literal, Mapping, Sequence
    from typing_extensions import NotRequired, TypedDict, Union, List


# tool call function
class ToolCallFunction(TypedDict):
    """
    Tool call function.
    """

    name: str
    "Name of the function."

    arguments: NotRequired[Mapping[str, Any]]
    "Arguments of the function."


# tool call
class ToolCall(TypedDict):
    """
    Model tool calls.
    """

    function: ToolCallFunction
    "Function to be called."


# message
class Message(TypedDict):
    """
    Chat message.
    """

    role: Literal["user", "assistant", "system", "tool"]
    "Assumed role of the message. Response messages always has role 'assistant' or 'tool'."

    content: NotRequired[str]
    "Content of the message. Response messages contains message fragments when streaming."

    images: NotRequired[Sequence[Any]]
    """
  Optional list of image data for multimodal models.

  Valid input types are:

  - `str` or path-like object: path to image file
  - `bytes` or bytes-like object: raw image data

  Valid image formats depend on the model. See the model card for more information.
  """

    tool_calls: NotRequired[Sequence[ToolCall]]
    """
  Tools calls to be made by the model.
  """



# xnano
# hammad saeed // 2024

__all__ = [
    # lib level resource
    # exported at top level
    "console",
    # ----------------------------------------
    # agents
    # ----------------------------------------
    "Agent",
    "create_agent",
    "Steps",
    # ----------------------------------------
    # completions
    # ----------------------------------------
    "completion",
    "async_completion",
    # ----------------------------------------
    # data
    # ----------------------------------------
    "Database",
    "VectorStore",
    # ----------------------------------------
    # generators
    # ----------------------------------------
    "generate_code",
    "generate_function",
    "generate_classification",
    "async_generate_classification",
    "generate_chunks",
    "generate_extraction",
    "async_generate_extraction",
    "generate_sql",
    "generate_system_prompt",
    "generate_qa_pairs",
    "generate_answers",
    "generate_questions",
    "generate_validation",
    "async_generate_validation",
    # multimodal
    "multimodal_generate_image",
    "multimodal_generate_audio",
    "multimodal_generate_transcription",
    # ----------------------------------------
    # models
    # ----------------------------------------
    "GenerativeModel",
    "model_patch",
    "model_unpatch",
    # ----------------------------------------
    # utils
    # ----------------------------------------
    "convert_yaml_to_pydantic",
    "format_messages",
    "repair_messages",
    "verify_messages_integrity",
    "swap_system_prompt",
    "convert_to_openai_tool",
    "MessageUtils",
    "Tools",
    # ----------------------------------------
    # text
    # ----------------------------------------
    "text_reader",
    "text_embeddings",
    "text_chunker",
]

# ----------------------------------------
# lib level resources
# ----------------------------------------

from .lib import console

# ----------------------------------------
# agents
# ----------------------------------------

from .agents import Agent, create_agent, Steps

# ----------------------------------------
# completions
# ----------------------------------------

from .completions import completion, async_completion

# ----------------------------------------
# data
# ----------------------------------------

from .data import Database, VectorStore

# ----------------------------------------
# generators
# ----------------------------------------

from .generators import (
    generate_code,
    generate_function,
    generate_classification,
    async_generate_classification,
    generate_chunks,
    generate_extraction,
    async_generate_extraction,
    generate_sql,
    generate_system_prompt,
    generate_qa_pairs,
    generate_answers,
    generate_questions,
    generate_validation,
    async_generate_validation,
    # multimodal
    multimodal_generate_image,
    multimodal_generate_audio,
    multimodal_generate_transcription,
)

# ----------------------------------------
# models
# ----------------------------------------

from .models import GenerativeModel, model_patch, model_unpatch

# ----------------------------------------
# utils
# ----------------------------------------

from .resources.utils._routing import (
    convert_yaml_to_pydantic,
    format_messages,
    repair_messages,
    verify_messages_integrity,
    swap_system_prompt,
    convert_to_openai_tool,
    MessageUtils,
    Tools,
)

# ----------------------------------------
# text
# ----------------------------------------

from .text import text_reader, text_embeddings, text_chunker



__all__ = [
    "convert_yaml_to_pydantic",
    "format_messages",
    "repair_messages",
    "verify_messages_integrity",
    "swap_system_prompt",
    "MessageUtils",
    "convert_to_openai_tool",
    "Tools",
]

from .convert_yaml_to_pydantic import convert_yaml_to_pydantic
from .messages import (
    format_messages,
    repair_messages,
    verify_messages_integrity,
    swap_system_prompt,
    Messages as MessageUtils,
)
from ..completions.resources.tool_calling import convert_to_openai_tool
from .tools import Tools



from instructor import from_litellm, Mode

# interal imports
from ...lib import console, XNANOException
from .resources import messages as message_utils
from .resources import structured_outputs, tool_calling, utils, batch

from ...types.completions.arguments import CompletionArguments
from ...types.embeddings.memory import Memory
from ...types.completions.params import (
    CompletionMessagesParam,
    CompletionAudioParam,
    CompletionChatModelsParam,
    CompletionContextParam,
    CompletionInstructorModeParam,
    CompletionResponseModelParam,
    CompletionModalityParam,
    CompletionPredictionContentParam,
    CompletionToolChoiceParam,
    CompletionToolsParam,
)
from ...types.completions.responses import (
    Response,
    StreamingResponse,
    AsyncStreamingResponse,
)

from pydantic import BaseModel

import httpx
import json
from typing import List, Optional, Union, overload, Literal, Callable, Any


# base client
class Completions:
    """Base Completions Resource"""

    ## init
    # no client args aside from verbosity
    # client uses litellm function methods for completions,
    # not instantiated client
    def __init__(self, verbose: bool = False):
        """
        Initializes the base completions client

        Args:
            verbose (bool): Whether to print verbose output
        """

        self.verbose = verbose

        try:
            self.import_litellm_methods()
        except Exception as e:
            raise XNANOException(f"Failed to initialize litellm methods: {e}")

        if self.verbose:
            console.message(
                "✅ [green]Successfully initialized [bold]completions[/bold] resource[/green]"
            )

    # litellm setup method
    def import_litellm_methods(self) -> None:
        """
        Imports the litellm methods
        """
        import litellm
        from litellm import (
            completion,
            batch_completion,
            acompletion,
            batch_completion_models_all_responses,
        )

        # drop params
        litellm.drop_params = True
        litellm.modify_params = True

        # set methods
        self.litellm_completion = completion
        self.litellm_batch_completion = batch_completion
        self.litellm_acompletion = acompletion
        self.litellm_batch_completion_models_all_responses = (
            batch_completion_models_all_responses
        )

        self.instructor_completion = None
        self.instructor_acompletion = None

    # INSTRUCTOR PATCH METHODS
    # PATCHES COMPLETION / ACOMPLETION FOR SYNC/ASYNC INSTRUCTOR CLIENT
    # instructor patch
    def instructor_patch(
        self, mode: CompletionInstructorModeParam = "tool_call"
    ) -> None:
        """
        Patches the completion methods with instructor mode
        CompletionInstructorModeParam is a literal list of the string values available in instructor.Mode

        Args:
            mode (CompletionInstructorModeParam): The mode to patch the completion methods with

        Returns:
            None
        """

        if mode is None:
            mode = "tool_call"

        # sync
        if not self.instructor_completion:
            self.instructor_completion = from_litellm(
                self.litellm_completion, mode=Mode(mode)
            )

        if self.verbose:
            console.message(
                f"✅ [green]Successfully patched synchronous completion methods w/ [bold]instructor[/bold] mode: [bold white]{mode if mode is not None else 'tool_call'}[/bold white][/green]"
            )

    def instructor_apatch(
        self, mode: CompletionInstructorModeParam = "tool_call"
    ) -> None:
        """
        Patches the async completion methods with instructor mode

        Args:
            mode (InstructorMode): The mode to patch the completion methods with

        Returns:
            None
        """

        if not mode:
            mode = "tool_call"

        if not self.instructor_acompletion:
            self.instructor_acompletion = from_litellm(
                self.litellm_acompletion, mode=Mode(mode)
            )

        if self.verbose:
            console.message(
                f"✅ [green]Successfully patched asynchronous completion methods w/ [bold]instructor[/bold] mode: [bold white]{mode}[/bold white][/green]"
            )

    # ------------------------------------------------------------
    # helpers
    # ------------------------------------------------------------
    def _handle_stream(self, response) -> StreamingResponse:
        """
        Handles the stream response
        """

        return response

    def _type_response_list(self, response: Response) -> List[Response]:
        """
        Types the response as a list of responses
        """

        return response

    # ------------------------------------------------------------
    # non - batch completion methods
    # ------------------------------------------------------------
    def _run_completion(self, args: CompletionArguments) -> Response:
        """
        Runs a completion
        """

        if args.response_model is not None:
            try:
                self.instructor_patch(mode=args.mode)
            except Exception as e:
                raise XNANOException(f"Failed to patch instructor: {e}")

            try:
                if not args.stream:
                    return self.instructor_completion.chat.completions.create(
                        **utils.build_post_request(args, instructor=True)
                    )
                else:
                    return self.instructor_completion.chat.completions.create_partial(
                        **utils.build_post_request(args, instructor=True)
                    )

            except Exception as e:
                raise XNANOException(f"Failed to run instructor completion: {e}")

        else:
            try:
                return self.litellm_completion(**utils.build_post_request(args))

            except Exception as e:
                raise XNANOException(f"Failed to run completion: {e}")

    async def _arun_completion(self, args: CompletionArguments) -> Response:
        """
        Runs a completion

        Args:
            args (CompletionArguments): The arguments to run the completion with

        Returns:
            CompletionResponse: The completion response
        """

        if args.response_model is not None:
            try:
                self.instructor_apatch(mode=args.mode)
            except Exception as e:
                raise XNANOException(f"Failed to patch instructor: {e}")

            try:
                if not args.stream:
                    return await self.instructor_acompletion.chat.completions.create(
                        **utils.build_post_request(args, instructor=True)
                    )
                else:
                    return await self.instructor_acompletion.chat.completions.create_partial(
                        **utils.build_post_request(args, instructor=True)
                    )

            except Exception as e:
                raise XNANOException(f"Failed to run instructor completion: {e}")

        else:
            try:
                return await self.litellm_acompletion(**utils.build_post_request(args))

            except Exception as e:
                raise XNANOException(f"Failed to run completion: {e}")

    # ------------------------------------------------------------
    # batch completion handler
    # ------------------------------------------------------------
    def _run_batch_completion(self, args: CompletionArguments):
        pass

    # ------------------------------------------------------------
    # base completion handler
    # ------------------------------------------------------------
    def run_completion(
        self,
        messages: CompletionMessagesParam,
        model: CompletionChatModelsParam = "gpt-4o-mini",
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        response_format: Optional[CompletionResponseModelParam] = None,
        tools: Optional[List[CompletionToolsParam]] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        n: Optional[int] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        # openai v1.0+ new params
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        # soon to be deprecated params by OpenAI
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        # set api_base, api_version, api_key
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        return_messages: Optional[bool] = None,
    ) -> Union[Response, StreamingResponse]:
        """
        Runs a completion

        Returns:
            CompletionResponse: The completion response
        """

        if instructor_mode:
            mode = instructor_mode

        if mode:
            self.instructor_patch(mode)

        # setup response
        responses = []

        # flags
        ran_tools = False
        is_batch_completion = False
        is_tool_execution = False
        embedding_context_string = None
        original_response_model = (
            response_model if response_model or response_format else None
        )

        # set flags
        if tools and run_tools is True:
            is_tool_execution = True

        # ------------------------------------------------------------
        # batch handling
        # ------------------------------------------------------------

        # get batch determination
        batch_determination = batch.determine_batch_needed(messages, model)

        if self.verbose:
            console.message(
                f"✅ [green]Batch determination: [bold white]{batch_determination}[/bold white][/green]"
            )

        if batch_determination.multi_model_batch_job:
            # Get local variables as dict and remove unwanted keys
            local_vars = locals()
            local_vars.pop("self", None)
            local_vars.pop("batch_determination", None)
            local_vars.pop("responses", None)
            local_vars.pop("ran_tools", None)
            local_vars.pop("is_batch_completion", None)
            local_vars.pop("is_tool_execution", None)
            local_vars.pop("embedding_context_string", None)
            local_vars.pop("original_response_model", None)
            local_vars.pop("response_model", None)
            return batch.create_batch_completion_job_with_multiple_models(
                **local_vars, response_model=response_model
            )

        if batch_determination.multi_message_batch_job:
            # Get local variables as dict and remove unwanted keys
            local_vars = locals()
            local_vars.pop("self", None)
            local_vars.pop("batch_determination", None)
            local_vars.pop("responses", None)
            local_vars.pop("ran_tools", None)
            local_vars.pop("is_batch_completion", None)
            local_vars.pop("is_tool_execution", None)
            local_vars.pop("embedding_context_string", None)
            local_vars.pop("original_response_model", None)
            local_vars.pop("response_model", None)
            return batch.create_batch_completion_job(
                **local_vars, response_model=response_model
            )

        # ------------------------------------------------------------
        # structured output handling
        # ------------------------------------------------------------

        embeddings = memory
        embeddings_limit = memory_limit

        if embeddings:
            if not isinstance(embeddings, list):
                embeddings = [embeddings]

            for embedding in embeddings:
                embedding_context = embedding._build_context(
                    messages=messages,
                    model=model,
                    api_key=api_key,
                    base_url=base_url,
                    organization=organization,
                    limit=embeddings_limit,
                )

            embedding_context_string = "\n\n".join(embedding_context)

        if context:
            if embedding_context_string:
                context = f"{embedding_context_string}\n\n{context}"

        # set response_format as response_model if given instead of response_model
        if response_format:
            console.warning(
                "'response_format' is an allowed argument, but it is preferred to use 'response_model' instead"
            )
            response_model = response_format

        # handle response model
        if response_model:
            response_model = structured_outputs.handle_response_model(response_model)

        try:
            # format messages
            messages = message_utils.format_messages(messages)

            # add context to messages
            if context:
                messages = message_utils.add_context_to_messages(messages, context)

            if self.verbose:
                if is_batch_completion is False:
                    console.message(
                        f"✅ [green]Successfully formatted {len(messages)} messages[/green]"
                    )
                else:
                    console.message(
                        f"✅ [green]Successfully formatted batch of {len(messages)} messages[/green]"
                    )

        except Exception as e:
            raise XNANOException(
                f"Failed to validate messages, please ensure they are formatted correctly: {e}"
            )

        # build args
        try:
            args = utils.collect_completion_args(locals())
        except Exception as e:
            raise XNANOException(f"Failed to build completion arguments: {e}")

        # format tools
        if tools:
            try:
                args.tools = []

                for tool in tools:
                    if isinstance(tool, str):
                        try:
                            if self.verbose:
                                console.message(
                                    "String tool detected, checking for prebuilt tool..."
                                )

                            args.tools.append(
                                tool_calling.generate_tool(self, tool, model)
                            )

                        except Exception as e:
                            raise XNANOException(
                                f"Failed to generate tool named: {tool}: {e}"
                            )
                    else:
                        args.tools.append(tool_calling.convert_to_tool(tool))
            except Exception as e:
                raise XNANOException(f"Failed to format tools: {e}")

        # handle batch
        if is_batch_completion:
            try:
                # TODO: implement
                # tool execution warning for batch completions
                if is_tool_execution:
                    console.warning(
                        "Tool execution is not supported for batch completions yet."
                    )

                # run & return
                return self._run_batch_completion(args)

            except Exception as e:
                raise XNANOException(f"Failed to run batch completion: {e}")

        # handle non batch
        else:
            try:
                if response_model and tools:
                    args.response_model = None

                response = self._run_completion(args)
            except Exception as e:
                raise XNANOException(f"Failed to run completion: {e}")

        # handle tool execution
        if tools and run_tools is True:
            args.messages.append(response.choices[0].message.model_dump())
            responses.append(response)

            # detect tool calls
            if response.choices[0].message.tool_calls:
                # flag
                ran_tools = False

                # iterate through tool calls
                for tool_call in response.choices[0].message.tool_calls:
                    # check if tool name is in args.tools
                    if tool_call.function.name in [tool.name for tool in args.tools]:
                        try:
                            for tool in args.tools:
                                if tool.name == tool_call.function.name and callable(
                                    tool.function
                                ):
                                    output = tool.function(
                                        **json.loads(tool_call.function.arguments)
                                    )

                                    if output is None:
                                        console.warning(
                                            "Did your function return an output? No output was returned from the tool."
                                        )

                                        output = "Tool executed successfully!"
                                    ran_tools = True

                                    args.messages.append(
                                        {
                                            "role": "tool",
                                            "content": json.dumps(f"{str(output)}"),
                                            "tool_call_id": tool_call.id,
                                        }
                                    )

                                    responses.append(
                                        {
                                            "role": "tool",
                                            "content": json.dumps(f"{str(output)}"),
                                            "tool_call_id": tool_call.id,
                                        }
                                    )

                                    if self.verbose:
                                        console.message(
                                            f"✅ [green]Successfully ran tool: [bold white]{tool_call.function.name}[/bold white] with args: [bold white]{tool_call.function.arguments}[/bold white][/green]"
                                        )

                        except Exception as e:
                            raise XNANOException(
                                f"Failed to run tool: {e} with args: {tool_call.function.arguments}"
                            )

                    # return if none
                    else:
                        return response

            # return if no tools ran
            if not ran_tools:
                return response

            # get final response
            # add response model if given
            if response_model:
                args.response_model = response_model
                args.tools = None
                args.tool_choice = None
                args.parallel_tool_calls = None

                # NOTE: HACK!
                # instructor does not support tool calls in the final response model
                # so we have to make nice
                args.messages = structured_outputs.make_nice_with_instructor(
                    args.messages
                )
            # run & return
            try:
                response = self._run_completion(args)

                responses.append(response)
            except Exception as e:
                raise XNANOException(f"Failed to run final completion: {e}")

        if response_model:
            if isinstance(original_response_model, type) and not issubclass(
                original_response_model, BaseModel
            ):
                if return_messages:
                    if self.verbose:
                        console.message("✅ [green]Returning messages[/green]")

                    return responses

                else:
                    return response.response

            else:
                from ..models.mixin import patch

                if self.verbose:
                    console.message("✅ [green]Successfully patched response[/green]")

                return patch(response)

        if return_messages:
            if responses:
                return responses
            else:
                return response
        else:
            return response

    # ASYNC
    async def arun_completion(
        self,
        messages: CompletionMessagesParam,
        model: CompletionChatModelsParam = "gpt-4o-mini",
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        response_format: Optional[CompletionResponseModelParam] = None,
        tools: Optional[List[CompletionToolsParam]] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        n: Optional[int] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        # openai v1.0+ new params
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        # soon to be deprecated params by OpenAI
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        # set api_base, api_version, api_key
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        return_messages: Optional[bool] = None,
    ) -> Union[Response, AsyncStreamingResponse]:
        """
        Runs an async completion

        Returns:
            Union[Response, StreamingResponse]: The completion response
        """

        if instructor_mode:
            mode = instructor_mode

        if mode:
            self.instructor_patch(mode)

        # setup response
        responses = []

        # flags
        ran_tools = False
        is_batch_completion = False
        is_tool_execution = False
        embedding_context_string = None

        # set flags
        if isinstance(messages, list) and isinstance(messages[0], list):
            is_batch_completion = True
        if tools and run_tools is True:
            is_tool_execution = True

        # ------------------------------------------------------------
        # structured output handling
        # ------------------------------------------------------------

        embeddings = memory
        embeddings_limit = memory_limit

        if embeddings:
            if not isinstance(embeddings, list):
                embeddings = [embeddings]

            for embedding in embeddings:
                embedding_context = embedding._build_context(
                    messages=messages,
                    model=model,
                    api_key=api_key,
                    base_url=base_url,
                    organization=organization,
                    limit=embeddings_limit,
                )

            embedding_context_string = "\n\n".join(embedding_context)

        if context:
            if embedding_context_string:
                context = f"{embedding_context_string}\n\n{context}"

        # set response_format as response_model if given instead of response_model
        if response_format:
            console.warning(
                "'response_format' is an allowed argument, but it is preferred to use 'response_model' instead"
            )
            response_model = response_format

        # handle response model
        if response_model:
            response_model = structured_outputs.handle_response_model(response_model)

        try:
            # format messages
            messages = message_utils.format_messages(messages)

            # add context to messages
            if context:
                messages = message_utils.add_context_to_messages(messages, context)

            if self.verbose:
                if is_batch_completion is False:
                    console.message(
                        f"✅ [green]Successfully formatted {len(messages)} messages[/green]"
                    )
                else:
                    console.message(
                        f"✅ [green]Successfully formatted batch of {len(messages)} messages[/green]"
                    )

        except Exception as e:
            raise XNANOException(
                f"Failed to validate messages, please ensure they are formatted correctly: {e}"
            )

        # build args
        try:
            args = utils.collect_completion_args(locals())
        except Exception as e:
            raise XNANOException(f"Failed to build completion arguments: {e}")

        # format tools
        if tools:
            try:
                args.tools = []

                for tool in tools:
                    if isinstance(tool, str):
                        try:
                            if self.verbose:
                                console.message(
                                    "String tool detected, checking for prebuilt tool..."
                                )

                            args.tools.append(
                                tool_calling.generate_tool(self, tool, model)
                            )

                        except Exception as e:
                            raise XNANOException(
                                f"Failed to generate tool named: {tool}: {e}"
                            )
                    else:
                        args.tools.append(tool_calling.convert_to_tool(tool))
            except Exception as e:
                raise XNANOException(f"Failed to format tools: {e}")

        # handle batch
        if is_batch_completion:
            try:
                # TODO: implement
                # tool execution warning for batch completions
                if is_tool_execution:
                    console.warning(
                        "Tool execution is not supported for batch completions yet."
                    )

                # run & return
                return self._run_batch_completion(args)

            except Exception as e:
                raise XNANOException(f"Failed to run batch completion: {e}")

        # handle non batch
        else:
            try:
                if response_model and tools:
                    args.response_model = None

                response = await self._arun_completion(args)
            except Exception as e:
                raise XNANOException(f"Failed to run completion: {e}")

        # handle tool execution
        if tools and run_tools is True:
            args.messages.append(response.choices[0].message.model_dump())
            responses.append(response)

            # detect tool calls
            if response.choices[0].message.tool_calls:
                # flag
                ran_tools = False

                # iterate through tool calls
                for tool_call in response.choices[0].message.tool_calls:
                    # check if tool name is in args.tools
                    if tool_call.function.name in [tool.name for tool in args.tools]:
                        try:
                            for tool in args.tools:
                                if tool.name == tool_call.function.name and callable(
                                    tool.function
                                ):
                                    output = tool.function(
                                        **json.loads(tool_call.function.arguments)
                                    )

                                    if output is None:
                                        console.warning(
                                            "Did your function return an output? No output was returned from the tool."
                                        )

                                        output = "Tool executed successfully!"
                                    ran_tools = True

                                    args.messages.append(
                                        {
                                            "role": "tool",
                                            "content": json.dumps(f"{str(output)}"),
                                            "tool_call_id": tool_call.id,
                                        }
                                    )

                                    responses.append(
                                        {
                                            "role": "tool",
                                            "content": json.dumps(f"{str(output)}"),
                                            "tool_call_id": tool_call.id,
                                        }
                                    )

                                    if self.verbose:
                                        console.message(
                                            f"✅ [green]Successfully ran tool: [bold white]{tool_call.function.name}[/bold white] with args: [bold white]{tool_call.function.arguments}[/bold white][/green]"
                                        )

                        except Exception as e:
                            raise XNANOException(
                                f"Failed to run tool: {e} with args: {tool_call.function.arguments}"
                            )

                    # return if none
                    else:
                        return response

            # return if no tools ran
            if not ran_tools:
                return response

            # get final response
            # add response model if given
            if response_model:
                args.response_model = response_model
                args.tools = None
                args.tool_choice = None
                args.parallel_tool_calls = None

                # NOTE: HACK!
                # instructor does not support tool calls in the final response model
                # so we have to make nice
                args.messages = structured_outputs.make_nice_with_instructor(
                    args.messages
                )
            # run & return
            try:
                response = await self._arun_completion(args)

                responses.append(response)
            except Exception as e:
                raise XNANOException(f"Failed to run final completion: {e}")

        if return_messages:
            if responses:
                return responses
            else:
                return response
        else:
            return response

    # ------------------------------------------------------------
    # public
    # ------------------------------------------------------------
    def completion(
        self,
        # messages
        # if str, will be formatted as user message
        # if list of list of messages, will be sent as a batch request
        messages: CompletionMessagesParam,
        model: CompletionChatModelsParam = "gpt-4o-mini",
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        # instructor mode param now implemented but not fully built in
        mode: Optional[CompletionInstructorModeParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        response_format: Optional[CompletionResponseModelParam] = None,
        tools: Optional[List[CompletionToolsParam]] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        n: Optional[int] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        # openai v1.0+ new params
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        # soon to be deprecated params by OpenAI
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        # set api_base, api_version, api_key
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        return_messages: Optional[bool] = None,
    ) -> Union[Response, StreamingResponse]:
        """
        Create a chat completion or completion(s)

        Example:
        ```python
        completion(messages="hi", model="gpt-4o-mini")
        ```

        Args:
            messages (CompletionMessageParam): Messages to send to the model
            model (CompletionChatModelParam): Model to use for the completion
            context (CompletionContext): Context to use for the completion
            mode (CompletionInstructorMode): Instructor mode to use for the completion
            response_model (CompletionResponseModelParam): Response model to use for the completion
            response_format (CompletionResponseModelParam): Response format to use for the completion
            tools (List[CompletionToolType]): Tools to use for the completion
            run_tools (bool): Run tools for the completion
            tool_choice (CompletionToolChoiceParam): Tool choice to use for the completion
            parallel_tool_calls (bool): Parallel tool calls to use for the completion
            api_key (str): API key to use for the completion
            base_url (str): Base URL to use for the completion
            organization (str): Organization to use for the completion
            n (int): Number of completions to use for the completion
            timeout (Union[float, str, httpx.Timeout]): Timeout to use for the completion
            temperature (float): Temperature to use for the completion
            top_p (float): Top P to use for the completion
            stream_options (dict): Stream options to use for the completion
            stop (str): Stop to use for the completion
            max_completion_tokens (int): Max completion tokens to use for the completion
            max_tokens (int): Max tokens to use for the completion
            modalities (List[ChatCompletionModality]): Modalities to use for the completion
            prediction (ChatCompletionPredictionContentParam): Prediction to use for the completion
            audio (ChatCompletionAudioParam): Audio to use for the completion
            presence_penalty (float): Presence penalty to use for the completion
            frequency_penalty (float): Frequency penalty to use for the completion
            logit_bias (dict): Logit bias to use for the completion
            user (str): User to use for the completion
            seed (int): Seed to use for the completion
            logprobs (bool): Logprobs to use for the completion
            top_logprobs (int): Top logprobs to use for the completion
            deployment_id (str): Deployment ID to use for the completion
            extra_headers (dict): Extra headers to use for the completion
            functions (List): Functions to use for the completion
            function_call (str): Function call to use for the completion
            api_version (str): API version to use for the completion
            model_list (list): Model list to use for the completion
            stream (bool): Stream to use for the completion
            verbose (bool): Verbose to use for the completion
        """

        if instructor_mode:
            mode = instructor_mode

        # run completion
        return self.run_completion(
            messages=messages,
            model=model,
            context=context,
            memory=memory,
            memory_limit=memory_limit,
            mode=mode,
            response_model=response_model,
            response_format=response_format,
            tools=tools,
            run_tools=run_tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            api_key=api_key,
            base_url=base_url,
            organization=organization,
            n=n,
            stream=stream,
            timeout=timeout,
            temperature=temperature,
            top_p=top_p,
            stream_options=stream_options,
            stop=stop,
            max_completion_tokens=max_completion_tokens,
            max_tokens=max_tokens,
            modalities=modalities,
            prediction=prediction,
            audio=audio,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            user=user,
            seed=seed,
            logprobs=logprobs,
            top_logprobs=top_logprobs,
            deployment_id=deployment_id,
            extra_headers=extra_headers,
            functions=functions,
            function_call=function_call,
            api_version=api_version,
            model_list=model_list,
            return_messages=return_messages,
        )

    @staticmethod
    @overload
    def _completion(
        messages: CompletionMessagesParam,
        stream: Literal[True],
        model: CompletionChatModelsParam = "gpt-4o-mini",
        **kwargs,
    ) -> Union[StreamingResponse, List[StreamingResponse]]: ...

    @staticmethod
    @overload
    def _completion(
        messages: CompletionMessagesParam,
        model: CompletionChatModelsParam = "gpt-4o-mini",
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        response_format: Optional[CompletionResponseModelParam] = None,
        tools: Optional[List[CompletionToolsParam]] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        n: Optional[int] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Literal[False] = False,
        return_messages: Optional[bool] = None,
        verbose: Optional[bool] = None,
    ) -> Union[Response, List[Response]]: ...

    @staticmethod
    def _completion(
        messages: CompletionMessagesParam,
        model: CompletionChatModelsParam = "gpt-4o-mini",
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        response_format: Optional[CompletionResponseModelParam] = None,
        tools: Optional[List[CompletionToolsParam]] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        n: Optional[int] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: bool = False,
        return_messages: Optional[bool] = None,
        verbose: Optional[bool] = None,
    ) -> Union[Response, StreamingResponse, List[Response], List[StreamingResponse]]:
        # deprecated
        if mode:
            console.mark_deprecated(
                "xnano.completion() | `[bold red]mode[/bold red]` will soon be deprecated. Use `[bold green]instructor_mode[/bold green]` for setting the Instructor Generation Mode instead."
            )

        # deprecated
        if max_tokens:
            console.mark_deprecated(
                "xnano.completion() | `[bold red]max_tokens[/bold red]` is now deprecated. [bold sky_blue3]OpenAI[/bold sky_blue3] uses `[bold green]max_completion_tokens[/bold green]` instead."
            )
            max_completion_tokens = max_tokens

        local_args = locals()
        local_args.pop("verbose", None)

        try:
            return Completions(verbose=verbose).run_completion(**local_args)
        except Exception as e:
            raise XNANOException(f"Failed to run completion: {e}")

    # ------------------------------------------------------------
    # async
    # ------------------------------------------------------------
    async def acompletion(
        self,
        messages: CompletionMessagesParam,
        model: CompletionChatModelsParam = "gpt-4o-mini",
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        response_format: Optional[CompletionResponseModelParam] = None,
        tools: Optional[List[CompletionToolsParam]] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        n: Optional[int] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        # openai v1.0+ new params
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        # soon to be deprecated params by OpenAI
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        # set api_base, api_version, api_key
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        return_messages: Optional[bool] = None,
    ) -> Union[Response, AsyncStreamingResponse]:
        """
        Asynchronously create a chat completion or completion(s)

        Example:
        ```python
        completion(messages="hi", model="gpt-4o-mini")
        ```

        Args:
            messages (CompletionMessageParam): Messages to send to the model
            model (CompletionChatModelParam): Model to use for the completion
            context (CompletionContext): Context to use for the completion
            mode (CompletionInstructorMode): Instructor mode to use for the completion
            response_model (CompletionResponseModelParam): Response model to use for the completion
            response_format (CompletionResponseModelParam): Response format to use for the completion
            tools (List[CompletionToolType]): Tools to use for the completion
            run_tools (bool): Run tools for the completion
            tool_choice (CompletionToolChoiceParam): Tool choice to use for the completion
            parallel_tool_calls (bool): Parallel tool calls to use for the completion
            api_key (str): API key to use for the completion
            base_url (str): Base URL to use for the completion
            organization (str): Organization to use for the completion
            n (int): Number of completions to use for the completion
            timeout (Union[float, str, httpx.Timeout]): Timeout to use for the completion
            temperature (float): Temperature to use for the completion
            top_p (float): Top P to use for the completion
            stream_options (dict): Stream options to use for the completion
            stop (str): Stop to use for the completion
            max_completion_tokens (int): Max completion tokens to use for the completion
            max_tokens (int): Max tokens to use for the completion
            modalities (List[ChatCompletionModality]): Modalities to use for the completion
            prediction (ChatCompletionPredictionContentParam): Prediction to use for the completion
            audio (ChatCompletionAudioParam): Audio to use for the completion
            presence_penalty (float): Presence penalty to use for the completion
            frequency_penalty (float): Frequency penalty to use for the completion
            logit_bias (dict): Logit bias to use for the completion
            user (str): User to use for the completion
            seed (int): Seed to use for the completion
            logprobs (bool): Logprobs to use for the completion
            top_logprobs (int): Top logprobs to use for the completion
            deployment_id (str): Deployment ID to use for the completion
            extra_headers (dict): Extra headers to use for the completion
            functions (List): Functions to use for the completion
            function_call (str): Function call to use for the completion
            api_version (str): API version to use for the completion
            model_list (list): Model list to use for the completion
            stream (bool): Stream to use for the completion
        """

        if instructor_mode:
            mode = instructor_mode

        return await self.arun_completion(
            messages=messages,
            model=model,
            context=context,
            memory=memory,
            memory_limit=memory_limit,
            mode=mode,
            response_model=response_model,
            response_format=response_format,
            tools=tools,
            run_tools=run_tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            api_key=api_key,
            base_url=base_url,
            organization=organization,
            n=n,
            stream=stream,
            timeout=timeout,
            temperature=temperature,
            top_p=top_p,
            stream_options=stream_options,
            stop=stop,
            max_completion_tokens=max_completion_tokens,
            max_tokens=max_tokens,
            modalities=modalities,
            prediction=prediction,
            audio=audio,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            user=user,
            seed=seed,
            logprobs=logprobs,
            top_logprobs=top_logprobs,
            deployment_id=deployment_id,
            extra_headers=extra_headers,
            functions=functions,
            function_call=function_call,
            api_version=api_version,
            model_list=model_list,
            return_messages=return_messages,
        )


@overload
def completion(
    messages: CompletionMessagesParam,
    stream: Literal[True],
    model: CompletionChatModelsParam = "gpt-4o-mini",
    **kwargs,
) -> Union[StreamingResponse, List[StreamingResponse]]: ...


@overload
def completion(
    messages: CompletionMessagesParam,
    model: CompletionChatModelsParam = "gpt-4o-mini",
    context: Optional[CompletionContextParam] = None,
    memory: Optional[Union[Memory, List[Memory]]] = None,
    memory_limit: Optional[int] = None,
    instructor_mode: Optional[CompletionInstructorModeParam] = None,
    mode: Optional[CompletionInstructorModeParam] = None,
    response_model: Optional[CompletionResponseModelParam] = None,
    response_format: Optional[CompletionResponseModelParam] = None,
    tools: Optional[List[CompletionToolsParam]] = None,
    run_tools: Optional[bool] = None,
    tool_choice: Optional[CompletionToolChoiceParam] = None,
    parallel_tool_calls: Optional[bool] = None,
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    organization: Optional[str] = None,
    n: Optional[int] = None,
    timeout: Optional[Union[float, str, httpx.Timeout]] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None,
    stream_options: Optional[dict] = None,
    stop=None,
    max_completion_tokens: Optional[int] = None,
    max_tokens: Optional[int] = None,
    modalities: Optional[List[CompletionModalityParam]] = None,
    prediction: Optional[CompletionPredictionContentParam] = None,
    audio: Optional[CompletionAudioParam] = None,
    presence_penalty: Optional[float] = None,
    frequency_penalty: Optional[float] = None,
    logit_bias: Optional[dict] = None,
    user: Optional[str] = None,
    seed: Optional[int] = None,
    logprobs: Optional[bool] = None,
    top_logprobs: Optional[int] = None,
    deployment_id=None,
    extra_headers: Optional[dict] = None,
    functions: Optional[List] = None,
    function_call: Optional[str] = None,
    api_version: Optional[str] = None,
    model_list: Optional[list] = None,
    stream: Literal[False] = False,
    return_messages: Optional[bool] = None,
    verbose: Optional[bool] = None,
) -> Union[Response, List[Response]]: ...


def completion(
    messages: CompletionMessagesParam,
    model: CompletionChatModelsParam = "gpt-4o-mini",
    context: Optional[CompletionContextParam] = None,
    memory: Optional[Union[Memory, List[Memory]]] = None,
    memory_limit: Optional[int] = None,
    instructor_mode: Optional[CompletionInstructorModeParam] = None,
    mode: Optional[CompletionInstructorModeParam] = None,
    response_model: Optional[CompletionResponseModelParam] = None,
    response_format: Optional[CompletionResponseModelParam] = None,
    tools: Optional[List[CompletionToolsParam]] = None,
    run_tools: Optional[bool] = None,
    tool_choice: Optional[CompletionToolChoiceParam] = None,
    parallel_tool_calls: Optional[bool] = None,
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    organization: Optional[str] = None,
    n: Optional[int] = None,
    timeout: Optional[Union[float, str, httpx.Timeout]] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None,
    stream_options: Optional[dict] = None,
    stop=None,
    max_completion_tokens: Optional[int] = None,
    max_tokens: Optional[int] = None,
    modalities: Optional[List[CompletionModalityParam]] = None,
    prediction: Optional[CompletionPredictionContentParam] = None,
    audio: Optional[CompletionAudioParam] = None,
    presence_penalty: Optional[float] = None,
    frequency_penalty: Optional[float] = None,
    logit_bias: Optional[dict] = None,
    user: Optional[str] = None,
    seed: Optional[int] = None,
    logprobs: Optional[bool] = None,
    top_logprobs: Optional[int] = None,
    deployment_id=None,
    extra_headers: Optional[dict] = None,
    functions: Optional[List] = None,
    function_call: Optional[str] = None,
    api_version: Optional[str] = None,
    model_list: Optional[list] = None,
    stream: bool = False,
    return_messages: Optional[bool] = None,
    verbose: Optional[bool] = None,
) -> Union[Response, StreamingResponse, List[Response], List[StreamingResponse]]:
    # deprecated
    if mode:
        console.mark_deprecated(
            "xnano.completion() | `[bold red]mode[/bold red]` will soon be deprecated. Use `[bold green]instructor_mode[/bold green]` for setting the Instructor Generation Mode instead."
        )

    # deprecated
    if max_tokens:
        console.mark_deprecated(
            "xnano.completion() | `[bold red]max_tokens[/bold red]` is now deprecated. [bold sky_blue3]OpenAI[/bold sky_blue3] uses `[bold green]max_completion_tokens[/bold green]` instead."
        )
        max_completion_tokens = max_tokens

    local_args = locals()
    local_args.pop("verbose", None)

    try:
        return Completions(verbose=verbose).run_completion(**local_args)
    except Exception as e:
        raise XNANOException(f"Failed to run completion: {e}")


@overload
def async_completion(
    messages: CompletionMessagesParam,
    stream: Literal[True],
    model: CompletionChatModelsParam = "gpt-4o-mini",
    **kwargs,
) -> Union[AsyncStreamingResponse, List[AsyncStreamingResponse]]: ...


@overload
def async_completion(
    messages: CompletionMessagesParam,
    model: CompletionChatModelsParam = "gpt-4o-mini",
    context: Optional[CompletionContextParam] = None,
    memory: Optional[Union[Memory, List[Memory]]] = None,
    memory_limit: Optional[int] = None,
    instructor_mode: Optional[CompletionInstructorModeParam] = None,
    mode: Optional[CompletionInstructorModeParam] = None,
    response_model: Optional[CompletionResponseModelParam] = None,
    response_format: Optional[CompletionResponseModelParam] = None,
    tools: Optional[List[CompletionToolsParam]] = None,
    run_tools: Optional[bool] = None,
    tool_choice: Optional[CompletionToolChoiceParam] = None,
    parallel_tool_calls: Optional[bool] = None,
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    organization: Optional[str] = None,
    n: Optional[int] = None,
    timeout: Optional[Union[float, str, httpx.Timeout]] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None,
    stream_options: Optional[dict] = None,
    stop=None,
    max_completion_tokens: Optional[int] = None,
    max_tokens: Optional[int] = None,
    modalities: Optional[List[CompletionModalityParam]] = None,
    prediction: Optional[CompletionPredictionContentParam] = None,
    audio: Optional[CompletionAudioParam] = None,
    presence_penalty: Optional[float] = None,
    frequency_penalty: Optional[float] = None,
    logit_bias: Optional[dict] = None,
    user: Optional[str] = None,
    seed: Optional[int] = None,
    logprobs: Optional[bool] = None,
    top_logprobs: Optional[int] = None,
    deployment_id=None,
    extra_headers: Optional[dict] = None,
    functions: Optional[List] = None,
    function_call: Optional[str] = None,
    api_version: Optional[str] = None,
    model_list: Optional[list] = None,
    stream: Literal[False] = False,
    return_messages: Optional[bool] = None,
    verbose: Optional[bool] = None,
) -> Union[Response, List[Response]]: ...


async def async_completion(
    messages: CompletionMessagesParam,
    model: CompletionChatModelsParam = "gpt-4o-mini",
    context: Optional[CompletionContextParam] = None,
    memory: Optional[Union[Memory, List[Memory]]] = None,
    memory_limit: Optional[int] = None,
    instructor_mode: Optional[CompletionInstructorModeParam] = None,
    mode: Optional[CompletionInstructorModeParam] = None,
    response_model: Optional[CompletionResponseModelParam] = None,
    response_format: Optional[CompletionResponseModelParam] = None,
    tools: Optional[List[CompletionToolsParam]] = None,
    run_tools: Optional[bool] = None,
    tool_choice: Optional[CompletionToolChoiceParam] = None,
    parallel_tool_calls: Optional[bool] = None,
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    organization: Optional[str] = None,
    n: Optional[int] = None,
    timeout: Optional[Union[float, str, httpx.Timeout]] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None,
    stream_options: Optional[dict] = None,
    stop=None,
    max_completion_tokens: Optional[int] = None,
    max_tokens: Optional[int] = None,
    modalities: Optional[List[CompletionModalityParam]] = None,
    prediction: Optional[CompletionPredictionContentParam] = None,
    audio: Optional[CompletionAudioParam] = None,
    presence_penalty: Optional[float] = None,
    frequency_penalty: Optional[float] = None,
    logit_bias: Optional[dict] = None,
    user: Optional[str] = None,
    # openai v1.0+ new params
    seed: Optional[int] = None,
    logprobs: Optional[bool] = None,
    top_logprobs: Optional[int] = None,
    deployment_id=None,
    extra_headers: Optional[dict] = None,
    # soon to be deprecated params by OpenAI
    functions: Optional[List] = None,
    function_call: Optional[str] = None,
    # set api_base, api_version, api_key
    api_version: Optional[str] = None,
    model_list: Optional[list] = None,
    stream: Optional[bool] = None,
    return_messages: Optional[bool] = None,
    verbose: Optional[bool] = None,
) -> Union[
    Response, AsyncStreamingResponse, List[Response], List[AsyncStreamingResponse]
]:
    """
    Asynchronously create a chat completion or completion(s)

    Example:
    ```python
    completion(messages="hi", model="gpt-4o-mini")
    ```

    Args:
        messages (CompletionMessageParam): Messages to send to the model
        model (CompletionChatModelParam): Model to use for the completion
        context (CompletionContext): Context to use for the completion
        mode (CompletionInstructorMode): Instructor mode to use for the completion
        response_model (CompletionResponseModelParam): Response model to use for the completion
        response_format (CompletionResponseModelParam): Response format to use for the completion
        tools (List[CompletionToolType]): Tools to use for the completion
        run_tools (bool): Run tools for the completion
        tool_choice (CompletionToolChoiceParam): Tool choice to use for the completion
        parallel_tool_calls (bool): Parallel tool calls to use for the completion
        api_key (str): API key to use for the completion
        base_url (str): Base URL to use for the completion
        organization (str): Organization to use for the completion
        n (int): Number of completions to use for the completion
        timeout (Union[float, str, httpx.Timeout]): Timeout to use for the completion
        temperature (float): Temperature to use for the completion
        top_p (float): Top P to use for the completion
        stream_options (dict): Stream options to use for the completion
        stop (str): Stop to use for the completion
        max_completion_tokens (int): Max completion tokens to use for the completion
        max_tokens (int): Max tokens to use for the completion
        modalities (List[ChatCompletionModality]): Modalities to use for the completion
        prediction (ChatCompletionPredictionContentParam): Prediction to use for the completion
        audio (ChatCompletionAudioParam): Audio to use for the completion
        presence_penalty (float): Presence penalty to use for the completion
        frequency_penalty (float): Frequency penalty to use for the completion
        logit_bias (dict): Logit bias to use for the completion
        user (str): User to use for the completion
        seed (int): Seed to use for the completion
        logprobs (bool): Logprobs to use for the completion
        top_logprobs (int): Top logprobs to use for the completion
        deployment_id (str): Deployment ID to use for the completion
        extra_headers (dict): Extra headers to use for the completion
        functions (List): Functions to use for the completion
        function_call (str): Function call to use for the completion
        api_version (str): API version to use for the completion
        model_list (list): Model list to use for the completion
        stream (bool): Stream to use for the completion
        verbose (bool): Verbose to use for the completion
    """

    # deprecated
    if mode:
        console.mark_deprecated(
            "xnano.async_completion() | `[bold red]mode[/bold red]` will soon be deprecated. Use `[bold green]instructor_mode[/bold green]` for setting the Instructor Generation Mode instead."
        )

    if max_tokens:
        console.mark_deprecated(
            "xnano.async_completion() | `[bold red]max_tokens[/bold red]` is now deprecated. [bold sky_blue3]OpenAI[/bold sky_blue3] uses `[bold green]max_completion_tokens[/bold green]` instead."
        )
        max_completion_tokens = max_tokens

    local_args = locals()
    local_args.pop("verbose", None)

    try:
        return await Completions(verbose=verbose).arun_completion(**local_args)
    except Exception as e:
        raise XNANOException(f"Failed to run completion: {e}")



from ...types.agents.agent_model import AgentModel
from ...types.agents.state import State
from ...types.agents.agent_response import AgentResponse
from ...types.completions.params import (
    CompletionMessagesParam,
    CompletionChatModelsParam,
    CompletionInstructorModeParam,
    CompletionResponseModelParam,
    CompletionToolChoiceParam,
    CompletionToolsParam,
)
from ...types.embeddings.memory import Memory
from .step import Steps

# - external ----------------------------------------------------
import json
from pydantic import BaseModel, create_model
from typing import Dict, List, Literal, Optional, Union, Type

Agent = Type["Agent"]
AgentResources = Type["AgentResources"]

class Agent:
    """
    base class for all agents in the xnano library.

    capabilities:
        - 'on-message' based - built to integrate both as a chatbot or an autogen style agent
        - multi-agent collaboration - ability to chat & collaborate on workflows with multiple agents
        - memory - uses xnano's vector store module for long term memory & summarization for short term memory
        - 'workflows' - use pydantic models like tools to create workflows with steps
        - 'steps' - like workflows but with much-much more fine-grained control over each step
    """

    state: State
    config: AgentModel
    resources: AgentResources

    def __init__(
        self,
        role: str = "assistant",
        name: Optional[str] = None,
        instructions: Optional[str] = None,
        planning: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
        summarization_steps: Optional[int] = 5,
        memory: Optional[List[Memory]] = None,
        model: Union[CompletionChatModelsParam, str] = "openai/gpt-4o-mini",
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        messages: Optional[CompletionMessagesParam] = None,
        verbose: bool = False,
        agents: Optional[List[Agent]] = None,
    ) -> None:
        """
        Initializes an agent with the given parameters.

        Example:
            ```python
            agent = Agent(role="user", name="my_agent")
            ```

            or

            ```python
            agent = Agent(
                role="user",
                name="my_agent",
                instructions="You are a helpful assistant",
                planning=True,
            )
            ```

        Args:
            role (str): The role of the agent.
            name (Optional[str]): The name of the agent.
            instructions (Optional[str]): The instructions for the agent.
            planning (Optional[bool]): Whether the agent should plan its actions.
            workflows (Optional[List[BaseModel]]): A list of pydantic models to use as workflows.
            summarization_steps (Optional[int]): The number of steps to use for summarization.
            memory (Optional[List[Memory]]): A list of memories to use for the agent.
            model (Union[CompletionChatModelsParam, str]): The model to use for the agent.
            instructor_mode (Optional[CompletionInstructorModeParam]): The instructor mode to use for the agent.
            base_url (Optional[str]): The base URL for the agent.
            api_key (Optional[str]): The API key for the agent.
            organization (Optional[str]): The organization for the agent.
            messages (Optional[CompletionMessagesParam]): The messages for the agent.
            verbose (bool): Whether to print verbose output.
            agents (Optional[List[Agent]]): A list of agents to collaborate with.
        """
        ...

    # ----------------------------------------------------------
    # - HELPERS
    # ----------------------------------------------------------

    def add_messages_to_state(self, messages: CompletionMessagesParam) -> None:
        """
        Persists messages to the agent's internal state message thread

        Args:
            messages (CompletionMessagesParam): The messages to persist

        Returns:
            None
        """
        ...

    def get_messages_from_state(self) -> CompletionMessagesParam:
        """
        Retrieves messages from the agent's internal state message thread

        Example:
            ```python
            messages = agent.get_messages_from_state()
            ```

        Returns:
            CompletionMessagesParam: The messages from the agent's internal state message thread
        """
        ...

    def add_response_to_state(
        self,
        response: AgentResponse,
        instructor: bool = False,
    ) -> None:
        """
        Persists a response to the agent's internal state

        Example:
            ```python
            agent.add_response_to_state(response)
            ```

        Args:
            response (AgentResponse): The response to persist
            instructor (bool): Whether the response is an instructor response

        Returns:
            None
        """
        ...

    def get_system_prompt(
        self,
        tools: Optional[List[BaseModel]] = None,
    ) -> str:
        """
        Retrieves the system prompt for the agent

        Example:
            ```python
            system_prompt = agent.get_system_prompt()
            ```

        Args:
            tools (Optional[List[BaseModel]]): The tools to use for the system prompt

        Returns:
            str: The system prompt for the agent
        """
        ...

    # ----------------------------------------------------------
    # - SUMMARIZATION
    # ----------------------------------------------------------

    def build_summary(
        self,
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
    ) -> str:
        """
        Builds a summary of the agent's internal state message thread &
        adds it to the summary thread.

        **Note**: This method will remove the original messages from the message thread.

        Args:
            model (Optional[Union[CompletionChatModelsParam, str]]): The model to use for the summary
            api_key (Optional[str]): The API key to use for the summary
            base_url (Optional[str]): The base URL to use for the summary
            organization (Optional[str]): The organization to use for the summary

        Returns:
            str: The summary of the agent's internal state message thread
        """
        ...

    # ----------------------------------------------------------
    # - WORKFLOWS
    # ----------------------------------------------------------

    def get_workflows(
        self,
        workflows: Optional[List[BaseModel]] = None,
    ) -> List[BaseModel]:
        """
        Retrieves the workflows for the agent, along with any additional
        workflows passed in as an argument.

        Args:
            workflows (Optional[List[BaseModel]]): The additional workflows to use

        Returns:
            List[BaseModel]: The workflows for the agent
        """
        ...

    def workflow(
        self,
        workflow: BaseModel,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        # completion specific params
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
    ) -> AgentResponse:
        """
        Executes a specified workflow directly, without running workflow selection pipeline.

        Example:
            ```python
            from pydantic import BaseModel, Field

            class MyWorkflow(BaseModel):
                write_code: str = Field(description="Write a web scraper in python")
                format_code : str

            response = agent.workflow(MyWorkflow)

            print(response.workflow)
            ```

        Args:
            workflow (BaseModel): The workflow to execute
            messages (Optional[CompletionMessagesParam]): The messages to use for the workflow
            agents (Optional[List[Agent]]): The agents to collaborate with
            model (Optional[Union[CompletionChatModelsParam, str]]): The model to use for the workflow
            base_url (Optional[str]): The base URL to use for the workflow
            api_key (Optional[str]): The API key to use for the workflow
            organization (Optional[str]): The organization to use for the workflow
            tools (Optional[CompletionToolsParam]): The tools to use for the workflow
            instructor_mode (Optional[CompletionInstructorModeParam]): The instructor mode to use for the workflow
            response_model (Optional[CompletionResponseModelParam]): The response model to use for the workflow
            tool_choice (Optional[CompletionToolChoiceParam]): The tool choice to use for the workflow
            parallel_tool_calls (Optional[bool]): Whether to use parallel tool calls for the workflow

        Returns:
            AgentResponse: The response from the workflow
        """
        ...

    # ----------------------------------------------------------
    # - COMPLETIONS
    # ----------------------------------------------------------

    def completion(
        self,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        workflows: Optional[List[BaseModel]] = None,
        # completion specific params
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
    ) -> AgentResponse:
        """
        Runs an agent chat 'completion' & saves user inputs, workflows & assistant responses to the agent's internal state.

        ## Notes:

        ### Response Format:
        - The response will be in a standard chat completion format.
        - If a workflow is ran, the response will contain a `response.workflow` attribute that contains the results of the workflow.

        ### Workflows:
        - If an agent has been given workflows, it will use a workflow selection pipeline to determine which workflow to run.
        - If a workflow is selected, it will be fully executed before the completion response is returned.

        ### Tools:
        - Currently all python functions are automatically executed.
        - This framework encourages the use of generating pydantic models for the arguments of your tools, and executing them that way.
            - This provides the ability of querying for individual arguments, or even multiple tools at once.
            - This also gives the benefit of all arguments being properly typed & validated.

        ### Multi-Agent Collaboration
        - If multiple agents are given, or helper agents are defined at init level; it will run a multi-agent collaborative completion.
        - All agents will be given the opportunity to run, and will be allowed to see the entire message history at the time of their turn.
        - If nested agents have workflows, they will also run their workflow selection pipeline.

        # Examples:
            **Standard Completion:**
            ```python
            response = agent.completion(messages="Hello, how are you?")
            ```

            **Workflow Execution:**
            ```python
            response = agent.completion(workflow=MyWorkflow)
            ```

            **Multi-Agent Collaboration:**
            ```python
            response = agent.completion(agents=[agent1, agent2])
            ```

            **Getting Responses:**
            ```python
            # to print response content
            print(response.choices[0].message.content)

            # to get the workflow results
            print(response.workflow)
            ```

        Args:
            messages (Optional[CompletionMessagesParam]): The messages to use for the completion
            agents (Optional[List[Agent]]): The agents to collaborate with
            workflows (Optional[List[BaseModel]]): The workflows to execute
            model (Optional[Union[CompletionChatModelsParam, str]]): The model to use for the completion
            base_url (Optional[str]): The base URL to use for the completion
            api_key (Optional[str]): The API key to use for the completion
            organization (Optional[str]): The organization to use for the completion
            tools (Optional[CompletionToolsParam]): The tools to use for the completion
            instructor_mode (Optional[CompletionInstructorModeParam]): The instructor mode to use for the completion
            response_model (Optional[CompletionResponseModelParam]): The response model to use for the completion
            tool_choice (Optional[CompletionToolChoiceParam]): The tool choice to use for the completion
            parallel_tool_calls (Optional[bool]): Whether to use parallel tool calls for the completion

        Returns:
            AgentResponse: The response from the completion
        """
        ...

    def run_completion(
        self,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        # completion specific params
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
    ) -> AgentResponse:
        """
        Runs an agent chat 'completion'. Does not save to the agent's internal state.

        ## Notes:

        ### Response Format:
        - The response will be in a standard chat completion format.
        - If a workflow is ran, the response will contain a `response.workflow` attribute that contains the results of the workflow.

        ### Workflows:
        - If an agent has been given workflows, it will use a workflow selection pipeline to determine which workflow to run.
        - If a workflow is selected, it will be fully executed before the completion response is returned.

        ### Tools:
        - Currently all python functions are automatically executed.
        - This framework encourages the use of generating pydantic models for the arguments of your tools, and executing them that way.
            - This provides the ability of querying for individual arguments, or even multiple tools at once.
            - This also gives the benefit of all arguments being properly typed & validated.

        ### Multi-Agent Collaboration
        - If multiple agents are given, or helper agents are defined at init level; it will run a multi-agent collaborative completion.
        - All agents will be given the opportunity to run, and will be allowed to see the entire message history at the time of their turn.
        - If nested agents have workflows, they will also run their workflow selection pipeline.

        # Examples:
            **Standard Completion:**
            ```python
            response = agent.completion(messages="Hello, how are you?")
            ```

            **Workflow Execution:**
            ```python
            response = agent.completion(workflow=MyWorkflow)
            ```

            **Multi-Agent Collaboration:**
            ```python
            response = agent.completion(agents=[agent1, agent2])
            ```

            **Getting Responses:**
            ```python
            # to print response content
            print(response.choices[0].message.content)

            # to get the workflow results
            print(response.workflow)
            ```

        Args:
            messages (Optional[CompletionMessagesParam]): The messages to use for the completion
            agents (Optional[List[Agent]]): The agents to collaborate with
            workflows (Optional[List[BaseModel]]): The workflows to execute
            model (Optional[Union[CompletionChatModelsParam, str]]): The model to use for the completion
            base_url (Optional[str]): The base URL to use for the completion
            api_key (Optional[str]): The API key to use for the completion
            organization (Optional[str]): The organization to use for the completion
            tools (Optional[CompletionToolsParam]): The tools to use for the completion
            instructor_mode (Optional[CompletionInstructorModeParam]): The instructor mode to use for the completion
            response_model (Optional[CompletionResponseModelParam]): The response model to use for the completion
            tool_choice (Optional[CompletionToolChoiceParam]): The tool choice to use for the completion
            parallel_tool_calls (Optional[bool]): Whether to use parallel tool calls for the completion

        Returns:
            AgentResponse: The response from the completion
        """
        ...

    # ----------------------------------------------------------
    # - STEPS
    # ----------------------------------------------------------

    def steps(self) -> Steps:
        """
        Create a new step execution handler

        Example:
            ```python
            steps = agent.steps()

            @steps.step("collect_data")
            def collect_data(agent: Agent, state: StepState, input_data: Dict):
                response = agent.completion(
                    messages="Collect relevant data for analysis"
                )
                return {"data": response.choices[0].message.content}

            @steps.step("analyze_data", depends_on=["collect_data"])
            def analyze_data(agent: Agent, state: StepState, input_data: Dict):
                data = input_data["collect_data"]["data"]
                response = agent.completion(
                    messages=f"Analyze this data: {data}"
                )
                return {"analysis": response.choices[0].message.content}
            ```

        Args:
            None

        Returns:
            Steps: A new step execution handler
        """
        ...

    # ----------------------------------------------------------
    # - COLLABORATION
    # ----------------------------------------------------------

    def collaborate(
        self,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        max_steps: int = 5,
        # completion specific params
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
    ) -> AgentResponse:
        """
        Conducts a task-oriented conversation until completion or max steps reached.
        Maintains thread state and integrates with existing agent capabilities.

        Example:
            ```python
            response = agent.conversation(
                messages="Write a web scraper in python",
                agents=[agent1, agent2],
                max_steps=10,
            )
            ```

        Args:
            messages (Optional[CompletionMessagesParam]): The messages to use for the conversation
            agents (Optional[List[Agent]]): The agents to collaborate with
            max_steps (int): The maximum number of steps to run
            model (Optional[Union[CompletionChatModelsParam, str]]): The model to use for the conversation
            base_url (Optional[str]): The base URL to use for the conversation
            api_key (Optional[str]): The API key to use for the conversation
            organization (Optional[str]): The organization to use for the conversation
            tools (Optional[CompletionToolsParam]): The tools to use for the conversation
            instructor_mode (Optional[CompletionInstructorModeParam]): The instructor mode to use for the conversation
            response_model (Optional[CompletionResponseModelParam]): The response model to use for the conversation
            tool_choice (Optional[CompletionToolChoiceParam]): The tool choice to use for the conversation
            parallel_tool_calls (Optional[bool]): Whether to use parallel tool calls for the conversation
            workflows (Optional[List[BaseModel]]): The workflows to execute

        Returns:
            AgentResponse: The response from the conversation
        """
        ...

    def update_instructions(self, instructions: str):
        """
        Updates the agent's instructions

        Example:
            ```python
            agent.update_instructions("You are a helpful assistant")
            ```

        Args:
            instructions (str): The new instructions for the agent

        Returns:
            None
        """
        ...

    def reset_state(self):
        """
        Resets the agent's internal state

        Example:
            ```python
            agent.reset_state()
            ```
        """
        ...

    def plan(
        self,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
    ) -> List[AgentResponse]:
        """
        Generates a plan to accomplish the given task and executes it step by step.

        Args:
            messages (Optional[CompletionMessagesParam]): Initial messages or user request.
            agents (Optional[List[Agent]]): Other agents to collaborate with.
            model (Optional[Union[CompletionChatModelsParam, str]]): Model to use for completions.
            base_url (Optional[str]): Base URL for the API.
            api_key (Optional[str]): API key for authentication.
            organization (Optional[str]): Organization identifier.
            tools (Optional[CompletionToolsParam]): Tools available to the agent.
            instructor_mode (Optional[CompletionInstructorModeParam]): Instructor mode settings.
            response_model (Optional[CompletionResponseModelParam]): Expected response model.
            tool_choice (Optional[CompletionToolChoiceParam]): Tool choice parameters.
            parallel_tool_calls (Optional[bool]): Whether to run tool calls in parallel.
            workflows (Optional[List[BaseModel]]): Workflows available to the agent.

        Returns:
            List[AgentResponse]: The list of responses from executing each step.
        """
        ...

    def chat_completion(
        self,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        # completion specific params
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
    ) -> AgentResponse:
        """
        Runs a chat completion with the agent, without executing any workflows or multi-agent collaboration.

        Example:
            ```python
            response = agent.chat_completion(messages="Hello, how are you?")
            ```

        Args:
            messages (Optional[CompletionMessagesParam]): The messages to use for the chat completion
            agents (Optional[List[Agent]]): The agents to collaborate with
            model (Optional[Union[CompletionChatModelsParam, str]]): The model to use for the chat completion
            base_url (Optional[str]): The base URL to use for the chat completion
            api_key (Optional[str]): The API key to use for the chat completion
            organization (Optional[str]): The organization to use for the chat completion
            tools (Optional[CompletionToolsParam]): The tools to use for the chat completion
            instructor_mode (Optional[CompletionInstructorModeParam]): The instructor mode to use for the chat completion
            response_model (Optional[CompletionResponseModelParam]): The response model to use for the chat completion
            tool_choice (Optional[CompletionToolChoiceParam]): The tool choice to use for the chat completion
            parallel_tool_calls (Optional[bool]): Whether to use parallel tool calls for the chat completion
            workflows (Optional[List[BaseModel]]): The workflows to execute

        Returns:
            AgentResponse: The response from the chat completion
        """
        ...



from typing import List, Literal, Optional, Union
from pydantic import BaseModel, Field, create_model
from rich.progress import Progress, SpinnerColumn, TextColumn

from ..completions.main import completion
from ...lib import console
from ...types.embeddings.memory import Memory
from ...types.completions.params import (
    CompletionMessagesParam,
    CompletionAudioParam,
    CompletionChatModelsParam,
    CompletionContextParam,
    CompletionInstructorModeParam,
    CompletionResponseModelParam,
    CompletionModalityParam,
    CompletionPredictionContentParam,
    CompletionToolChoiceParam,
    CompletionToolsParam,
)
from ...types.completions.responses import Response
from ..completions.resources.messages import swap_system_prompt
from ...types.completions.messages import Message


PROMPT_TYPES = Literal["costar", "tidd-ec", "instruction", "reasoning"]


class Prompts:
    costar = """
## CONTEXT ##

You are a world class COSTAR prompt creator & engineer.
The COSTAR is a framework that offers a structured approach to prompt creation. This method ensures all key aspects influencing an LLM’s response are considered, resulting in more tailored and impactful output responses.

The COSTAR Framework Explained:
- Context : Providing background information helps the LLM understand the specific scenario.
- Objective (O): Clearly defining the task directs the LLM’s focus.
- Style (S): Specifying the desired writing style aligns the LLM response.
- Tone (T): Setting the tone ensures the response resonates with the required sentiment.
- Audience (A): Identifying the intended audience tailors the LLM’s response to be targeted to an audience.
- Response_Format (R): Providing the response format, like text or json, ensures the LLM outputs, and help build pipelines.

## INSTRUCTIONS ##

- Identify what the user is trying to achieve according to their latest message(s).
- Identify the context of the user's message(s).
- Generate the apprpriate block of text for the COSTAR system prompt, creating a character or persona, that will aid the LLM in achieving the user's objective.

Your response does not always have to include every component of the COSTAR framework, but your prompt should be well-writen, and easy for the LLM to understand.
    """
    tidd_ec = """
## CONTEXT ##

You are a world class Tidd-EC prompt creator & engineer.
The Tidd-EC is a framework that offers a structured approach to prompt creation. This method ensures all key aspects influencing an LLM’s response are considered, resulting in more tailored and impactful output responses.

The Tidd-EC Framework Explained:
- Task : Providing background information helps the LLM understand the specific scenario.
- Instructions : Clearly defining the task directs the LLM’s focus.
- Do : Specifying the desired writing style aligns the LLM response.
- Donts : Setting the tone ensures the response resonates with the required sentiment.

## INSTRUCTIONS ##

- Identify what the user is trying to achieve according to their latest message(s).
- Identify the context of the user's message(s).
- Generate the apprpriate block of text for the COSTAR system prompt, creating a character or persona, that will aid the LLM in achieving the user's objective.

Your response does not always have to include every component of the COSTAR framework, but your prompt should be well-writen, and easy for the LLM to understand.
"""
    instruction = """
## CONTEXT ##

You are a world-class instruction generator, specializing in creating clear, concise, and actionable bullet-point instructions.

## INSTRUCTIONS ##

- Analyze the user's request or task description.
- Generate a comprehensive set of bullet-point instructions that address the user's needs.
- Ensure each instruction is clear, specific, and actionable.
- Organize instructions in a logical sequence.
- Use concise language while maintaining clarity.
- Include any necessary warnings or precautions as separate bullet points.

Your response should be a well-structured list of instructions that guide the user through the task or process effectively.
    """
    reasoning = """
## CONTEXT ##

You are an expert in creating step-by-step reasoning processes, designed to break down complex problems or tasks into logical, sequential steps.

## INSTRUCTIONS ##

- Carefully analyze the user's problem or task.
- Create a structured list of reasoning steps or phases that address the problem.
- Ensure each step builds logically on the previous ones.
- Include explanatory notes for each step if necessary.
- Consider potential branching paths in the reasoning process.
- Conclude with a step that synthesizes the reasoning or reaches a conclusion.

Your response should be a clear, logical sequence of reasoning steps that guide through the problem-solving or decision-making process.
    """


class CostarSystemPrompt(BaseModel):
    context: Optional[str] = Field(
        ...,
        description="Provide background information to help the LLM understand the specific scenario. If given, must be minimum 4 sentences.",
    )
    objective: Optional[str] = Field(
        ...,
        description="The objective of the user defined task that the LLM is supposed to complete. If given, must be minimum 4 sentences.",
    )
    style: Optional[str] = Field(
        ...,
        description="The user-intended writing style of the response, intended to be produced by the LLM. If given, must be minimum 2 sentences.",
    )
    tone: Optional[str] = Field(
        ...,
        description="The tone of the response, something to help the response resonate with the required sentiment. If given, must be minimum 2 sentences.",
    )
    audience: Optional[str] = Field(
        ...,
        description="The intended audience of the response. If given, must be minimum 2 sentences.",
    )
    response_format: Optional[str] = Field(
        ...,
        description="Examples and desired outcome of the response. If given, must be minimum 2 sentences.",
    )


class TiddECSystemPrompt(BaseModel):
    task: Optional[str] = Field(
        ...,
        description="The description of the character the LLM is supposed to play, as well as the context of the scenario. If given, must be minimum 200 tokens.",
    )
    instructions: Optional[list[str]] = Field(
        ...,
        description="The instructions for the LLM to follow. Must be clear and consise. If given, must be minimum 200 tokens.",
    )
    do: Optional[list[str]] = Field(
        ...,
        description="Specifies actions the LLM should take to successfully complete the prompt. This includes the use of certain language, structures, or information that should be included in the response. Each item must be minimum 20 tokens.",
    )
    donts: Optional[list[str]] = Field(
        ...,
        description="Highlights actions or elements the LLM should avoid in its response. This is essential for preventing common errors or misinterpretations that could lead to inaccurate or irrelevant outputs. Each item must be minimum 20 tokens.",
    )
    examples: Optional[list[str]] = Field(
        ...,
        description="Provides concrete examples of desired outcomes or responses. This component is invaluable for guiding the LLM towards the expected format, style, or content of the response. Each item must be minimum 20 tokens.",
    )


class InstructionSystemPrompt(BaseModel):
    instructions: List[str] = Field(
        ...,
        description="A list of clear, concise, and actionable bullet-point instructions. Each instruction must be minimum 10 tokens.",
    )
    warnings: Optional[List[str]] = Field(
        None,
        description="Optional list of warnings or precautions related to the instructions. Each warning must be minimum 10 tokens.",
    )


class ReasoningSystemPrompt(BaseModel):
    steps: List[str] = Field(
        ...,
        description="A list of sequential reasoning steps or phases. Each step must be minimum 20 tokens.",
    )
    explanations: Optional[List[str]] = Field(
        None,
        description="Optional list of explanations for each step. Each explanation must be minimum 20 tokens.",
    )
    conclusion: str = Field(
        ...,
        description="A concluding statement or final step that synthesizes the reasoning process. Must be minimum 30 tokens.",
    )


PROMPT_TYPES_MAPPING = {
    "costar": CostarSystemPrompt,
    "tidd-ec": TiddECSystemPrompt,
    "instruction": InstructionSystemPrompt,
    "reasoning": ReasoningSystemPrompt,
}


def get_system_prompt(type: PROMPT_TYPES = "costar") -> dict[str, str]:
    prompt_content = getattr(Prompts, type.replace("-", "_"), None)
    if prompt_content is None:
        raise ValueError(f"Invalid prompt type: {type}")
    return {"role": "system", "content": prompt_content}


def get_response_model(type: PROMPT_TYPES = "costar") -> BaseModel:
    return PROMPT_TYPES_MAPPING[type]


def generate_system_prompt(
    instructions: Union[str, List[str]],
    type: PROMPT_TYPES = "costar",
    messages: Optional[List[Message]] = None,
    optimize: bool = False,
    process: Literal["sequential", "batch"] = "sequential",
    n: int = 1,
    batch_size: int = 3,
    model: Union[str, CompletionChatModelsParam] = "gpt-4o-mini",
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    organization: Optional[str] = None,
    temperature: Optional[float] = None,
    instructor_mode: CompletionInstructorModeParam = "tool_call",
    max_tokens: Optional[int] = None,
    max_completion_tokens: Optional[int] = None,
    response_format: Union[Literal["pydantic"], Literal["dict"], None] = None,
    progress_bar: Optional[bool] = True,
    verbose: bool = False,
) -> Union[BaseModel, str, dict, List[Union[BaseModel, str, dict]], List[Message]]:
    if verbose:
        console.message(f"Generating system prompt for {type}")
        console.message(f"Optimize: {optimize}")
        console.message(f"Process: {process}")
        console.message(f"Number of prompts: {n}")
        console.message(f"Batch size: {batch_size}")

    response_model = get_response_model(type=type)
    system_prompt = get_system_prompt(type=type)

    if isinstance(instructions, str):
        instructions = [instructions]

    if optimize:
        instructions = [
            f"Optimize the following system prompt:\n\n{instr}"
            for instr in instructions
        ]

    results = []

    if progress_bar:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            task_id = progress.add_task("Building Prompt...", total=len(instructions))

            if process == "sequential":
                for instr in instructions:
                    messages = [
                        system_prompt,
                        {
                            "role": "user",
                            "content": f"Generate a system prompt for the following instructions:\n\nINSTRUCTIONS:\n{instr}",
                        },
                    ]
                    if results:
                        messages.append(
                            {
                                "role": "assistant",
                                "content": f"Previously generated prompts:\n{results[-1]}",
                            }
                        )

                    result = completion(
                        messages=messages,
                        model=model,
                        response_model=response_model,
                        instructor_mode=instructor_mode,
                        max_tokens=max_tokens,
                        max_completion_tokens=max_completion_tokens,
                        temperature=temperature,
                        api_key=api_key,
                        base_url=base_url,
                        organization=organization,
                        verbose=verbose,
                    )
                    results.append(result)
                    progress.update(task_id, advance=1)
            else:  # batch process
                for i in range(0, len(instructions), batch_size):
                    batch = instructions[i : i + batch_size]
                    batch_message = (
                        "Generate system prompts for the following instructions:\n\n"
                    )
                    for idx, instr in enumerate(batch, 1):
                        batch_message += f"{idx}. {instr}\n\n"

                    response_model_batch = create_model(
                        "ResponseModel", items=(List[response_model], ...)
                    )

                    result = completion(
                        messages=[
                            system_prompt,
                            {"role": "user", "content": batch_message},
                        ],
                        model=model,
                        response_model=response_model_batch,
                        instructor_mode=instructor_mode,
                        max_tokens=max_tokens,
                        max_completion_tokens=max_completion_tokens,
                        temperature=temperature,
                        api_key=api_key,
                        base_url=base_url,
                        organization=organization,
                        verbose=verbose,
                    )

                    results.extend(result.items)
                    progress.update(task_id, advance=len(batch))
    else:
        if process == "sequential":
            for instr in instructions:
                messages = [
                    system_prompt,
                    {
                        "role": "user",
                        "content": f"Generate a system prompt for the following instructions:\n\nINSTRUCTIONS:\n{instr}",
                    },
                ]
                if results:
                    messages.append(
                        {
                            "role": "assistant",
                            "content": f"Previously generated prompts:\n{results[-1]}",
                        }
                    )

                result = completion(
                    messages=messages,
                    model=model,
                    response_model=response_model,
                    instructor_mode=instructor_mode,
                    max_tokens=max_tokens,
                    max_completion_tokens=max_completion_tokens,
                    temperature=temperature,
                    api_key=api_key,
                    base_url=base_url,
                    organization=organization,
                    verbose=verbose,
                )
                results.append(result)
        else:  # batch process
            for i in range(0, len(instructions), batch_size):
                batch = instructions[i : i + batch_size]
                batch_message = (
                    "Generate system prompts for the following instructions:\n\n"
                )
                for idx, instr in enumerate(batch, 1):
                    batch_message += f"{idx}. {instr}\n\n"

                response_model_batch = create_model(
                    "ResponseModel", items=(List[response_model], ...)
                )

                result = completion(
                    messages=[
                        system_prompt,
                        {"role": "user", "content": batch_message},
                    ],
                    model=model,
                    response_model=response_model_batch,
                    instructor_mode=instructor_mode,
                    max_tokens=max_tokens,
                    temperature=temperature,
                    api_key=api_key,
                    base_url=base_url,
                    organization=organization,
                    verbose=verbose,
                )

                results.extend(result.items)

    # Helper function to convert BaseModel to string
    def model_to_string(model_instance: BaseModel) -> str:
        response_string = []
        for field in model_instance.model_fields:
            value = getattr(model_instance, field, None)
            if isinstance(value, list):
                formatted_value = "\n".join(f"- {item}" for item in value)
            else:
                formatted_value = value
            response_string.append(f"## {field.capitalize()} ##\n{formatted_value}\n\n")
        return "\n".join(response_string)

    # Regenerate prompts to ensure second-person perspective
    regenerated_results = []
    for prompt in results:
        prompt_str = model_to_string(prompt)
        messages = [
            {
                "role": "system",
                "content": "Regenerate the following prompt to ensure it refers to the character in the second person (using 'you', 'your', etc.):\n\n"
                + prompt_str,
            }
        ]
        regenerated_prompt = completion(
            messages=messages,
            model=model,
            response_model=response_model,
            instructor_mode=instructor_mode,
            max_tokens=max_tokens,
            max_completion_tokens=max_completion_tokens,
            temperature=temperature,
            api_key=api_key,
            base_url=base_url,
            organization=organization,
            verbose=verbose,
        )
        regenerated_results.append(regenerated_prompt)

    # Format the regenerated results
    formatted_results = []
    for prompt in regenerated_results:
        formatted_results.append(model_to_string(prompt))

    # If messages were provided, swap the system prompt and return updated messages
    if messages is not None:
        result = (
            formatted_results[0] if len(formatted_results) == 1 else formatted_results
        )
        return swap_system_prompt(messages=messages, system_prompt=result)

    # Otherwise return the formatted results as before
    return formatted_results if len(formatted_results) > 1 else formatted_results[0]


if __name__ == "__main__":
    # Example usage
    instructions = [
        "Create a system prompt for a chatbot that helps users with programming questions.",
        "Generate a system prompt for an AI assistant that provides travel recommendations.",
    ]

    result = generate_system_prompt(
        instructions=instructions,
        type="costar",
        optimize=False,
        process="sequential",
        n=2,
        batch_size=2,
        verbose=True,
    )

    print("Generated System Prompts:")
    for idx, prompt in enumerate(result, 1):
        print(f"\nPrompt {idx}:")
        print(prompt)



from typing import List, Optional, Union, Any, Dict, Literal
from pydantic import BaseModel, create_model

from ...lib import console, XNANOException
from ..completions.main import completion
from ...types.completions.params import (
    CompletionChatModelsParam,
    CompletionInstructorModeParam,
)
from rich.progress import Progress, SpinnerColumn, TextColumn


class SQLQuery(BaseModel):
    query: str


class SQLQueryResult(BaseModel):
    query: str
    explanation: Optional[str] = None
    dialect: str
    parameters: Optional[Dict[str, Any]] = None


def generate_sql(
    input: Union[Dict, BaseModel, List[BaseModel]],
    objective: str,
    dialect: Literal["postgresql", "mysql", "sqlite", "mssql"] = "postgresql",
    include_explanation: bool = False,
    model: Union[str, CompletionChatModelsParam] = "gpt-4o-mini",
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    organization: Optional[str] = None,
    instructor_mode: CompletionInstructorModeParam = "tool_call",
    temperature: float = 0.2,
    progress_bar: Optional[bool] = True,
    verbose: bool = False,
) -> SQLQueryResult:
    """
    Generates an SQL query based on input data and objective.

    Examples:
        >>> data = {"users": [{"id": 1, "name": "John"}, {"id": 2, "name": "Jane"}]}
        >>> generate_sql(data, "Get all users where id > 1")
        SQLQueryResult(
            query="SELECT * FROM users WHERE id > 1",
            explanation="This query selects all users with ID greater than 1",
            dialect="postgresql"
        )

    Args:
        input: The input data structure (dict or Pydantic model(s))
        objective: The goal or objective of the SQL query
        dialect: SQL dialect to use
        include_explanation: Whether to include query explanation
        model: The model to use for generation
        api_key: Optional API key
        base_url: Optional base URL
        organization: Optional organization ID
        instructor_mode: Completion instructor mode
        temperature: Temperature for generation
        progress_bar: Whether to show progress bar
        verbose: Whether to show verbose output

    Returns:
        SQLQueryResult: Generated SQL query with optional explanation
    """

    if verbose:
        console.message(f"Generating SQL query for objective: {objective}")
        console.message(f"Using dialect: {dialect}")
        console.message(f"Input data type: {type(input)}")

    # Convert input data to a schema description
    if isinstance(input, dict):
        data_description = str(input)
    elif isinstance(input, list) and all(isinstance(x, BaseModel) for x in input):
        data_description = input[0].model_json_schema()
    elif isinstance(input, BaseModel):
        data_description = input.model_json_schema()
    else:
        raise XNANOException("Input data must be a dict, BaseModel, or List[BaseModel]")

    # Step 1: Generate the SQL query
    query_system_message = f"""
You are an expert SQL query generator. Generate a valid {dialect.upper()} SQL query based on the provided data structure and objective.

Data Structure:
{data_description}

Critical Instructions:
- Generate ONLY the SQL query text
- Follow {dialect.upper()} syntax and best practices
- Use appropriate table/column names from the provided data structure
- Ensure the query is optimized and efficient
- Do not include any DDL statements unless specifically requested
- Return only the query text, no explanation or additional text
"""

    if progress_bar:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            task = progress.add_task(
                "Generating SQL query...", total=2 if include_explanation else 1
            )

            query_result = completion(
                messages=[
                    {"role": "system", "content": query_system_message},
                    {"role": "user", "content": f"Objective: {objective}"},
                ],
                model=model,
                response_model=SQLQuery,
                instructor_mode=instructor_mode,
                temperature=temperature,
                api_key=api_key,
                base_url=base_url,
                organization=organization,
            )

            progress.advance(task)

            # Step 2: Generate explanation if requested
            explanation = None
            if include_explanation:
                explanation_system_message = """
You are an expert SQL analyst. Provide a clear, concise explanation of how the given SQL query works and what it accomplishes.

Instructions:
- Explain the query's purpose and functionality
- Highlight any important aspects or optimizations
- Keep the explanation clear and technical but accessible
- Focus on what the query does, not how to modify it
"""
                explanation_result = completion(
                    messages=[
                        {"role": "system", "content": explanation_system_message},
                        {"role": "user", "content": f"Query: {query_result.query}"},
                    ],
                    model=model,
                    response_model=create_model("Explanation", explanation=(str, ...)),
                    instructor_mode=instructor_mode,
                    temperature=temperature,
                    api_key=api_key,
                    base_url=base_url,
                    organization=organization,
                )
                explanation = explanation_result.explanation
                progress.advance(task)
    else:
        query_result = completion(
            messages=[
                {"role": "system", "content": query_system_message},
                {"role": "user", "content": f"Objective: {objective}"},
            ],
            model=model,
            response_model=SQLQuery,
            instructor_mode=instructor_mode,
            temperature=temperature,
            api_key=api_key,
            base_url=base_url,
            organization=organization,
        )

        # Generate explanation if requested
        explanation = None
        if include_explanation:
            explanation_system_message = """
You are an expert SQL analyst. Provide a clear, concise explanation of how the given SQL query works and what it accomplishes.

Instructions:
- Explain the query's purpose and functionality
- Highlight any important aspects or optimizations
- Keep the explanation clear and technical but accessible
- Focus on what the query does, not how to modify it
"""
            explanation_result = completion(
                messages=[
                    {"role": "system", "content": explanation_system_message},
                    {"role": "user", "content": f"Query: {query_result.query}"},
                ],
                model=model,
                response_model=create_model("Explanation", explanation=(str, ...)),
                instructor_mode=instructor_mode,
                temperature=temperature,
                api_key=api_key,
                base_url=base_url,
                organization=organization,
            )
            explanation = explanation_result.explanation

    return SQLQueryResult(
        query=query_result.query,
        explanation=explanation,
        dialect=dialect,
        parameters=None,  # Could be enhanced to detect and extract parameters
    )


if __name__ == "__main__":
    print(
        generate_sql(
            input={"users": [{"id": 1, "name": "John"}, {"id": 2, "name": "Jane"}]},
            objective="Get all users where id > 1",
        )
    )



# xnano . completions

__all__ = ["completion", "async_completion"]

# methods located at
# xnano.resources.completions.main
# this resource is the core of the library

from .resources.completions._routing import completion, async_completion



# agent initializer method

from . import helpers

# - types-------------------------------------------------------
from ...types.completions.params import (
    CompletionMessagesParam,
    CompletionChatModelsParam,
    CompletionInstructorModeParam,
    CompletionToolsParam,
)
from ...types.embeddings.memory import Memory

from .agent import Agent

from typing import Optional, List, Union
from pydantic import BaseModel


def create_agent(
    role: str = "Assistant",
    # optional
    name: Optional[str] = helpers.get_random_name(),
    instructions: Optional[str] = None,
    planning: Optional[bool] = False,
    workflows: Optional[List[BaseModel]] = None,
    summarization_steps: Optional[int] = 5,
    agents: Optional[List["Agent"]] = None,
    tools: Optional[List[CompletionToolsParam]] = None,
    # agent memory -- utilized differently than .completion(memory = ...)
    memory: Optional[List[Memory]] = None,
    # agent completion config params
    model: Union[CompletionChatModelsParam, str] = "openai/gpt-4o-mini",
    instructor_mode: Optional[CompletionInstructorModeParam] = None,
    base_url: Optional[str] = None,
    api_key: Optional[str] = None,
    organization: Optional[str] = None,
    messages: Optional[CompletionMessagesParam] = None,
    verbose: bool = False,
) -> Agent:
    """
    Initializes an agent with the given parameters.

    Example:
        ```python
        import xnano as x

        agent = x.create_agent(role="user", name="my_agent")
        ```

        or

        ```python
        agent = x.create_agent(
            role="user",
            name="my_agent",
            instructions="You are a helpful assistant",
            planning=True,
        )
        ```

    Args:
        role (str): The role of the agent.
        name (Optional[str]): The name of the agent.
        instructions (Optional[str]): The instructions for the agent.
        planning (Optional[bool]): Whether the agent should plan its actions.
        workflows (Optional[List[BaseModel]]): A list of pydantic models to use as workflows.
        summarization_steps (Optional[int]): The number of steps to use for summarization.
        memory (Optional[List[Memory]]): A list of memories to use for the agent.
        model (Union[CompletionChatModelsParam, str]): The model to use for the agent.
        instructor_mode (Optional[CompletionInstructorModeParam]): The instructor mode to use for the agent.
        base_url (Optional[str]): The base URL for the agent.
        api_key (Optional[str]): The API key for the agent.
        organization (Optional[str]): The organization for the agent.
        messages (Optional[CompletionMessagesParam]): The messages for the agent.
        verbose (bool): Whether to print verbose output.
        agents (Optional[List[Agent]]): A list of agents to collaborate with.
    """
    return Agent(
        role=role,
        name=name,
        instructions=instructions,
        planning=planning,
        workflows=workflows,
        summarization_steps=summarization_steps,
        memory=memory,
        model=model,
        instructor_mode=instructor_mode,
        base_url=base_url,
        api_key=api_key,
        organization=organization,
        messages=messages,
        tools=tools,
        verbose=verbose,
        agents=agents,
    )



from ..completions.main import completion
from pydantic import BaseModel, create_model
from typing import List, Optional, Union, Literal
from ..text.processing.text_chunker import text_chunker


class Chunk(BaseModel):
    text: str
    context: str


def generate_chunks(
    inputs: Union[str, List[str]],
    enhancement_type: Literal["summary", "context"] = "context",
    chunk_size: int = 512,
    model: str = "gpt-4o-mini",
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    progress_bar: bool = True,
) -> Union[List[str], List[List[str]]]:
    """
    Generates semantically enhanced chunks with either summaries or context strings.

    Args:
        inputs: Text(s) to chunk and enhance
        enhancement_type: Type of enhancement ("summary" or "context")
        chunk_size: Target size for each chunk in tokens
        model: LLM model to use for enhancement
        api_key: Optional API key for the LLM
        base_url: Optional base URL for API
        progress_bar: Show progress bar if True

    Returns:
        Enhanced chunks with context or summaries
    """
    # First, generate regular chunks
    raw_chunks = text_chunker(inputs, chunk_size=chunk_size, progress_bar=progress_bar)

    if isinstance(inputs, str):
        single_input = True
        chunks_list = [raw_chunks]
    else:
        single_input = False
        chunks_list = raw_chunks

    enhanced_results = []

    # System messages for different enhancement types
    system_messages = {
        "summary": """You are an expert at summarizing text chunks. Create a brief, informative summary that captures the key points of each chunk.""",
        "context": """You are an expert at understanding text context. For each chunk, generate a brief contextual description that helps understand its role in the larger document.""",
    }

    for chunks in chunks_list:
        enhanced_chunks = []

        for chunk in chunks:
            # Create prompt based on enhancement type
            if enhancement_type == "summary":
                user_message = (
                    f"Please provide a concise summary of this text chunk:\n\n{chunk}"
                )
            else:  # context
                user_message = f"Please provide a brief contextual description for this text chunk:\n\n{chunk}"

            # Use completion similar to classifier.py
            result = completion(
                messages=[
                    {"role": "system", "content": system_messages[enhancement_type]},
                    {"role": "user", "content": user_message},
                ],
                model=model,
                response_model=create_model("Enhancement", content=(str, ...)),
                mode="tool_call",
                api_key=api_key,
                base_url=base_url,
            )

            enhanced_chunks.append(Chunk(text=chunk, context=result.content))

        # Convert Chunk objects to strings before adding to results
        formatted_chunks = [
            f"""=== CHUNK ===
TEXT:
{chunk.text}

{enhancement_type.upper()}:
{chunk.context}
============="""
            for chunk in enhanced_chunks
        ]
        enhanced_results.append(formatted_chunks)

    return enhanced_results[0] if single_input else enhanced_results



from pydantic import BaseModel as PydanticBaseModel, create_model, Field
import httpx
from textwrap import dedent
from copy import deepcopy
from typing import (
    Type,
    TypeVar,
    Union,
    Dict,
    Any,
    Optional,
    List,
    overload,
)
from ...types.embeddings.memory import Memory
from ...types.completions.params import (
    CompletionMessagesParam,
    CompletionAudioParam,
    CompletionChatModelsParam,
    CompletionContextParam,
    CompletionInstructorModeParam,
    CompletionResponseModelParam,
    CompletionModalityParam,
    CompletionPredictionContentParam,
    CompletionToolChoiceParam,
    CompletionToolsParam,
)
from ...types.models.base_model_generation_process import BaseModelGenerationProcess
from ...types.models.mixin import BaseModelMixin as BaseModelMixinType
from pydantic import BaseModel as PydanticBaseModel
from ...types.completions._openai import ChatCompletion

class GenerativeModel(BaseModelMixinType, PydanticBaseModel):
    ...

    @overload
    async def model_async_completion(
        cls,
        messages: CompletionMessagesParam,
        model: CompletionChatModelsParam = "gpt-4o-mini",
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        response_format: Optional[CompletionResponseModelParam] = None,
        tools: Optional[List[CompletionToolsParam]] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        n: Optional[int] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        # openai v1.0+ new params
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        # soon to be deprecated params by OpenAI
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        # set api_base, api_version, api_key
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        return_messages: Optional[bool] = None,
        loader: Optional[bool] = True,
        verbose: Optional[bool] = None,
    ) -> Union[
        ChatCompletion,
        BaseModelMixinType,
        List[BaseModelMixinType],
        List[ChatCompletion],
    ]: ...
    @overload
    async def model_async_completion(
        self,
        messages: CompletionMessagesParam,
        model: CompletionChatModelsParam = "gpt-4o-mini",
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        response_format: Optional[CompletionResponseModelParam] = None,
        tools: Optional[List[CompletionToolsParam]] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        n: Optional[int] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        # openai v1.0+ new params
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        # soon to be deprecated params by OpenAI
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        # set api_base, api_version, api_key
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        return_messages: Optional[bool] = None,
        loader: Optional[bool] = True,
        verbose: Optional[bool] = None,
    ) -> Union[
        ChatCompletion,
        BaseModelMixinType,
        List[BaseModelMixinType],
        List[ChatCompletion],
    ]: ...
    @overload
    def model_completion(
        cls,
        messages: CompletionMessagesParam,
        model: CompletionChatModelsParam = "gpt-4o-mini",
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        response_format: Optional[CompletionResponseModelParam] = None,
        tools: Optional[List[CompletionToolsParam]] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        n: Optional[int] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        # openai v1.0+ new params
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        # soon to be deprecated params by OpenAI
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        # set api_base, api_version, api_key
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        return_messages: Optional[bool] = None,
        loader: Optional[bool] = True,
        verbose: Optional[bool] = None,
    ) -> Union[
        ChatCompletion,
        BaseModelMixinType,
        List[BaseModelMixinType],
        List[ChatCompletion],
    ]: ...
    @overload
    def model_completion(
        self,
        messages: CompletionMessagesParam,
        model: CompletionChatModelsParam = "gpt-4o-mini",
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        response_format: Optional[CompletionResponseModelParam] = None,
        tools: Optional[List[CompletionToolsParam]] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        n: Optional[int] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        # openai v1.0+ new params
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        # soon to be deprecated params by OpenAI
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        # set api_base, api_version, api_key
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        return_messages: Optional[bool] = None,
        loader: Optional[bool] = True,
        verbose: Optional[bool] = None,
    ) -> Union[
        ChatCompletion,
        BaseModelMixinType,
        List[BaseModelMixinType],
        List[ChatCompletion],
    ]: ...
    @overload
    def model_generate(
        cls,
        messages: CompletionMessagesParam = "",
        model: CompletionChatModelsParam = "gpt-4o-mini",
        process: BaseModelGenerationProcess = "batch",
        n: Optional[int] = 1,
        fields: Optional[List[str]] = None,
        regenerate: Optional[bool] = None,
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        tools: Optional[CompletionToolsParam] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        loader: Optional[bool] = True,
        verbose: Optional[bool] = None,
    ) -> Union[BaseModelMixinType, List[BaseModelMixinType]]: ...
    @overload
    def model_generate(
        self,
        messages: CompletionMessagesParam = "",
        model: CompletionChatModelsParam = "gpt-4o-mini",
        process: BaseModelGenerationProcess = "batch",
        n: Optional[int] = 1,
        fields: Optional[List[str]] = None,
        regenerate: Optional[bool] = None,
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        tools: Optional[CompletionToolsParam] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        loader: Optional[bool] = True,
        verbose: Optional[bool] = None,
    ) -> Union[BaseModelMixinType, List[BaseModelMixinType]]:
        ...
        """
        Generate or regenerate a full model or instance(s) of the model using LLM completion.
        Supports batching, multiple generations, specific field generation, and more.

        ## Examples:

        ### Generating a full model

        ```python
        from xnano import BaseModel

        class MyModel(BaseModel):
            name: str

        print(MyModel.model_generate(n=3))
        ```

        ### Generating a specific field

        ```python
        print(MyModel.model_generate(fields=["name"]))
        ```

        ### Regenerating a field

        ```python
        print(MyModel.model_generate(regenerate=True, fields=["name"]))
        ```

        Args:
            messages (CompletionMessagesParam): The messages to generate a completion for.
            model (CompletionChatModelsParam): The model to use for the completion.
            process (BaseModelGenerationProcess): The generation process to use for the completion.
            n (int): The number of completions to generate.
            fields (List[str]): The fields to generate.
            regenerate (bool): Whether to regenerate the fields.
            context (CompletionContextParam): The context to use for the completion.
            memory (Union[Memory, List[Memory]]): The memory to use for the completion.
            memory_limit (int): The memory limit to use for the completion.
            mode (CompletionInstructorModeParam): The mode to use for the completion.
            tools (CompletionToolsParam): The tools to use for the completion.
            run_tools (bool): Whether to run the tools.
            tool_choice (CompletionToolChoiceParam): The tool choice to use for the completion.
            parallel_tool_calls (bool): Whether to use parallel tool calls.
            api_key (str): The API key to use for the completion.
            base_url (str): The base URL to use for the completion.
            organization (str): The organization to use for the completion.
            timeout (Union[float, str, httpx.Timeout]): The timeout to use for the completion.
            temperature (float): The temperature to use for the completion.
            top_p (float): The top P to use for the completion.
            stream_options (dict): The stream options to use for the completion.
            stop (str): The stop sequence to use for the completion.
            max_completion_tokens (int): The maximum number of completion tokens to use for the completion.
            max_tokens (int): The maximum number of tokens to use for the completion.
            modalities (List[CompletionModalityParam]): The modalities to use for the completion.
            prediction (CompletionPredictionContentParam): The prediction content to use for the completion.
            audio (CompletionAudioParam): The audio to use for the completion.
            presence_penalty (float): The presence penalty to use for the completion.
            frequency_penalty (float): The frequency penalty to use for the completion.
            logit_bias (dict): The logit bias to use for the completion.
            user (str): The user to use for the completion.
            seed (int): The seed to use for the completion.
            logprobs (bool): Whether to return the log probabilities.
            top_logprobs (int): The top log probabilities to use for the completion.
            deployment_id (str): The deployment ID to use for the completion.
            extra_headers (dict): The extra headers to use for the completion.
            functions (List): The functions to use for the completion.
            function_call (str): The function call to use for the completion.
            api_version (str): The API version to use for the completion.
            model_list (list): The model list to use for the completion.
            stream (bool): Whether to stream the completion.
            loader (bool): Whether to show a loader.
            verbose (bool): Whether to show verbose output.
        """

    @overload
    async def model_async_generate(
        cls,
        messages: CompletionMessagesParam = "",
        model: CompletionChatModelsParam = "gpt-4o-mini",
        process: BaseModelGenerationProcess = "batch",
        n: Optional[int] = 1,
        fields: Optional[List[str]] = None,
        regenerate: Optional[bool] = None,
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        tools: Optional[CompletionToolsParam] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        loader: Optional[bool] = True,
        verbose: Optional[bool] = None,
    ) -> Union[BaseModelMixinType, List[BaseModelMixinType]]:
        ...
        """
        Generate or regenerate a full model or instance(s) of the model using LLM completion.
        Supports batching, multiple generations, specific field generation, and more.

        ## Examples:

        ### Generating a full model

        ```python
        from xnano import BaseModel

        class MyModel(BaseModel):
            name: str

        print(MyModel.model_generate(n=3))
        ```

        ### Generating a specific field

        ```python
        print(MyModel.model_generate(fields=["name"]))
        ```

        ### Regenerating a field

        ```python
        print(MyModel.model_generate(regenerate=True, fields=["name"]))
        ```

        Args:
            messages (CompletionMessagesParam): The messages to generate a completion for.
            model (CompletionChatModelsParam): The model to use for the completion.
            process (BaseModelGenerationProcess): The generation process to use for the completion.
            n (int): The number of completions to generate.
            fields (List[str]): The fields to generate.
            regenerate (bool): Whether to regenerate the fields.
            context (CompletionContextParam): The context to use for the completion.
            memory (Union[Memory, List[Memory]]): The memory to use for the completion.
            memory_limit (int): The memory limit to use for the completion.
            mode (CompletionInstructorModeParam): The mode to use for the completion.
            tools (CompletionToolsParam): The tools to use for the completion.
            run_tools (bool): Whether to run the tools.
            tool_choice (CompletionToolChoiceParam): The tool choice to use for the completion.
            parallel_tool_calls (bool): Whether to use parallel tool calls.
            api_key (str): The API key to use for the completion.
            base_url (str): The base URL to use for the completion.
            organization (str): The organization to use for the completion.
            timeout (Union[float, str, httpx.Timeout]): The timeout to use for the completion.
            temperature (float): The temperature to use for the completion.
            top_p (float): The top P to use for the completion.
            stream_options (dict): The stream options to use for the completion.
            stop (str): The stop sequence to use for the completion.
            max_completion_tokens (int): The maximum number of completion tokens to use for the completion.
            max_tokens (int): The maximum number of tokens to use for the completion.
            modalities (List[CompletionModalityParam]): The modalities to use for the completion.
            prediction (CompletionPredictionContentParam): The prediction content to use for the completion.
            audio (CompletionAudioParam): The audio to use for the completion.
            presence_penalty (float): The presence penalty to use for the completion.
            frequency_penalty (float): The frequency penalty to use for the completion.
            logit_bias (dict): The logit bias to use for the completion.
            user (str): The user to use for the completion.
            seed (int): The seed to use for the completion.
            logprobs (bool): Whether to return the log probabilities.
            top_logprobs (int): The top log probabilities to use for the completion.
            deployment_id (str): The deployment ID to use for the completion.
            extra_headers (dict): The extra headers to use for the completion.
            functions (List): The functions to use for the completion.
            function_call (str): The function call to use for the completion.
            api_version (str): The API version to use for the completion.
            model_list (list): The model list to use for the completion.
            stream (bool): Whether to stream the completion.
            loader (bool): Whether to show a loader.
            verbose (bool): Whether to show verbose output.
        """

    @overload
    async def model_async_generate(
        self,
        messages: CompletionMessagesParam = "",
        model: CompletionChatModelsParam = "gpt-4o-mini",
        process: BaseModelGenerationProcess = "batch",
        n: Optional[int] = 1,
        fields: Optional[List[str]] = None,
        regenerate: Optional[bool] = None,
        context: Optional[CompletionContextParam] = None,
        memory: Optional[Union[Memory, List[Memory]]] = None,
        memory_limit: Optional[int] = None,
        mode: Optional[CompletionInstructorModeParam] = None,
        tools: Optional[CompletionToolsParam] = None,
        run_tools: Optional[bool] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        timeout: Optional[Union[float, str, httpx.Timeout]] = None,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = None,
        stream_options: Optional[dict] = None,
        stop=None,
        max_completion_tokens: Optional[int] = None,
        max_tokens: Optional[int] = None,
        modalities: Optional[List[CompletionModalityParam]] = None,
        prediction: Optional[CompletionPredictionContentParam] = None,
        audio: Optional[CompletionAudioParam] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[dict] = None,
        user: Optional[str] = None,
        seed: Optional[int] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        deployment_id=None,
        extra_headers: Optional[dict] = None,
        functions: Optional[List] = None,
        function_call: Optional[str] = None,
        api_version: Optional[str] = None,
        model_list: Optional[list] = None,
        stream: Optional[bool] = None,
        loader: Optional[bool] = True,
        verbose: Optional[bool] = None,
    ) -> Union[BaseModelMixinType, List[BaseModelMixinType]]: ...

# -------------------------------------------------------------------------------------------------
# patch
# -------------------------------------------------------------------------------------------------

def patch(
    model: Union[Type[PydanticBaseModel], PydanticBaseModel],
) -> Union[Type[GenerativeModel], GenerativeModel]:
    """
    A function or decorator that patches a Pydantic BaseModel class or instance to build in xnano completions. Either
    run the function or decorate a class, or run the function on an instance.

    ### Examples:

    ```python
    from xnano import patch
    from pydantic import BaseModel

    @patch
    class MyModel(BaseModel):
        ...
    ```

    or

    ```python
    from xnano import patch
    from pydantic import BaseModel

    model = MyModel()
    patched_model = patch(model)
    ```

    Args:
        model: A Pydantic BaseModel class or instance.

    Returns:
        A patched Pydantic BaseModel class or instance.
    """
    ...

def unpatch(
    model: Union[
        Type[GenerativeModel],
        GenerativeModel,
        Type[PydanticBaseModel],
        PydanticBaseModel,
    ],
) -> Union[Type[PydanticBaseModel], PydanticBaseModel]:
    """
    A function that unpatches a patched Pydantic BaseModel class or instance.

    ### Examples:

    ```python
    from xnano import unpatch
    from pydantic import BaseModel

    @patch
    class MyModel(BaseModel):
        ...

    unpatched_model = unpatch(MyModel)
    ```

    Args:
        model: A patched Pydantic BaseModel class or instance.

    Returns:
        A Pydantic BaseModel class or instance.
    """
    ...



from ...types.agents.step import StepState
from typing import Dict, List, Optional, Any, Callable, Type

Agent = Type["Agent"]

class Steps:
    """
    execution handler for strict step-by-step workflows in xnano agents
    """

    def __init__(self, agent: Optional[Agent] = None, verbose: bool = False) -> None:
        """
        Initializes a steps workflow

        Args:
            agent (Optional[Agent]): The agent to use for the steps workflow
            verbose (bool): Whether to print verbose output
        """
        ...

    def step(
        self,
        name: str,
        depends_on: Optional[List[str]] = None,
        condition: Optional[Callable[[StepState], bool]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Callable:
        """
        Step decorator that can be used both with and without an agent

        Example:
            ```python
            @steps.step("process_data", depends_on=["fetch_data"])
            def process_data(state: StepState, input_data: Dict[str, Any]) -> Any:
                # Process the data
                return processed_data
            ```

        Args:
            name (str): The name of the step
            depends_on (Optional[List[str]]): The names of the steps that must be executed before this step
            condition (Optional[Callable[[StepState], bool]]): A condition that must be met for the step to be executed
            metadata (Optional[Dict[str, Any]]): Metadata for the step

        Returns:
            Callable: The decorated function
        """
        ...

    def add_step(
        self,
        name: str,
        handler: Callable,
        depends_on: Optional[List[str]] = None,
        condition: Optional[Callable[[StepState], bool]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Programmatically add a step without using the decorator

        Example:
            ```python
            steps.add_step("process_data", process_data_func, depends_on=["fetch_data"])
            ```

        Args:
            name (str): The name of the step
            handler (Callable): The function to execute for the step
            depends_on (Optional[List[str]]): The names of the steps that must be executed before this step
            condition (Optional[Callable[[StepState], bool]]): A condition that must be met for the step to be executed
            metadata (Optional[Dict[str, Any]]): Metadata for the step

        Returns:
            None
        """
        ...

    def execute(self) -> Dict[str, Any]:
        """
        Execute all steps in the correct order

        Example:
            ```python
            @steps.step("process_data", depends_on=["fetch_data"])
            def process_data(state: StepState, input_data: Dict[str, Any]) -> Any:
                # Process the data
                return processed_data

            results = steps.execute()
            ```

        Returns:
            Dict[str, Any]: Results from all completed steps
        """
        ...

    def reset(self) -> None:
        """Reset all steps to their initial state"""
        ...



# Agent Tests

from xnano import Agent
import pytest


def test_agents_blank_init():
    agent = Agent()

    assert agent.config.name is not None
    assert agent.config.role is not None
    assert agent.config.model is not None


# -------------------------------------------------------------------------------------------------
# Single Agent Completions (Non Workflow / Chaining)
# -------------------------------------------------------------------------------------------------


def test_agents_single_completion_context():
    agent = Agent(name="John")

    response = agent.completion("What is your name?")

    assert "john" in response.choices[0].message.content.lower()


def test_agents_completion_with_short_term_memory():
    agent = Agent(name="John")

    response = agent.completion("My name is Steve")

    response = agent.completion("What is my name?")

    assert "steve" in response.choices[0].message.content.lower()



# Generative Model Tests

from xnano import GenerativeModel, model_patch
from pydantic import BaseModel
from typing import Literal
import pytest


# Initializing Model Directly
def test_generative_model_init():
    class Person(GenerativeModel):
        name: str
        age: int

    person = Person(name="John", age=30)

    assert person.name == "John" and person.age == 30
    assert isinstance(person, Person) and issubclass(Person, BaseModel)


# Patching Pydantic Models
def test_generative_model_patch():
    @model_patch
    class Person(BaseModel):
        name: str
        age: int

    person = Person(name="John", age=30)

    assert person.name == "John" and person.age == 30


# -------------------------------------------------------------------------------------------------
# Generation Tests
# -------------------------------------------------------------------------------------------------


def test_generative_model_generate_single():
    class Information(GenerativeModel):
        instruction: str

    information = Information.model_generate()

    assert information.instruction is not None


def test_generative_model_generate_single_with_messages():
    class Information(GenerativeModel):
        instruction: str

    information = Information.model_generate(messages="How do i tie my shoes?")

    assert information.instruction is not None
    assert "shoe" in information.instruction.lower()


def test_generative_model_generate_multiple():
    class Person(GenerativeModel):
        name: str
        age: int

    people = Person.model_generate(n=2)

    assert len(people) == 2


def test_generative_model_field_regeneration():
    class Weather(GenerativeModel):
        condition: Literal["summer", "winter", "spring", "fall"]
        weather: str

    weather = Weather(condition="summer", weather="snowing")

    weather = weather.model_generate(
        messages="Please update the condition", fields=["condition"]
    )

    assert weather.condition != "summer"
    assert weather.condition in ["winter", "spring", "fall"]


# -------------------------------------------------------------------------------------------------
# Completion Tests
# -------------------------------------------------------------------------------------------------


def test_generative_model_completion():
    class Color(GenerativeModel):
        color: str

    color = Color(color="blue")

    response = color.model_completion(messages="What is my favorite color?")

    assert "blue" in response.choices[0].message.content.lower()
